static char rcsid[] = "@(#)$Id: read_rc.c,v 2.56 2013-10-13 07:18:52 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.56 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                       (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************
 *
 * Most of code copied from Elm 2.4 src/read_rc.c. It have following copyright:
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

#define SAVE_OPTS
#include "elm_defs.h"
#include "rc_imp.h"
#include "save_opts.h"
#include "s_elm.h"
#include "s_me.h"
#include "unidata.h"
#include "cs_imp.h"

#include "patchlevel.h"

#include <errno.h> 
#ifndef ANSI_C
extern int errno;
#endif

#include <limits.h>

DEBUG_VAR(Debug,__FILE__,"config");

static const unsigned char *cs2us P_((const char *str));
static const unsigned char *cs2us(str) 
     const char *str;
{
    return (const unsigned char *)str;
}

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

#ifdef PWDINSYS
#  include <sys/pwd.h>
#else
#  include <pwd.h>
#endif

#ifndef I_UNISTD 
char *getlogin();
unsigned short getgid(), getuid(); 
struct passwd *getpwuid();
#endif

# ifdef DOUNAME
#  include <sys/utsname.h>
# endif

struct hist_record {
    char *tag;
    struct rc_save_info_rec * entry;
    int changed;


    char  ** old_line;
    int   line_count;
};

static struct elmrc_recorder {
    struct hist_record *recs;
    int  rec_count;
    
} * recorders[RC_MODE_COUNT];

struct elmrc_recorder  * enable_recorder(mode)
     enum record_mode mode;
{
    if (mode < 0 || mode >= RC_MODE_COUNT) 
	panic("RC PANIC",__FILE__,__LINE__,"enable_recorder",
	      "Bad mode",0);

    if (! recorders[mode] ) {
	struct elmrc_recorder * X;

	X = safe_malloc(sizeof (*X));

	bzero((void *)X, sizeof (* (X)));

	X->recs      = NULL;
	X->rec_count = 0;

	recorders[mode] = X;
    }


    return recorders[mode];
}

static struct hist_record * record1 P_((struct elmrc_recorder  *recorder,
					struct rc_save_info_rec *X,
					char *tag));

static struct hist_record * record1(recorder,X,tag)
     struct elmrc_recorder  *recorder;
     struct rc_save_info_rec *X;
     char *tag;
{				       
    struct hist_record *r;


    recorder->recs = safe_realloc(recorder->recs,
					 sizeof (recorder->recs[0]) *
					 ( recorder->rec_count + 1 ));

    r = & (recorder->recs[recorder->rec_count]);

    bzero((void *)r, sizeof (*r));

    r->entry = X;
    r->tag   = NULL;

    if (tag)
	r->tag = safe_strdup(tag);

    r->old_line    = 0;
    r->line_count  = 0;
    r->changed     = 0;

    recorder->rec_count++;

    return r;
}

struct hist_record * rc_need_hist_record(mode,X,tag)
     enum record_mode mode;
     struct rc_save_info_rec * X;
     char *tag;
{
    int i;


    if (! recorders[mode] ) 
	return NULL;


    /* If old value is already recorded, 
       do not add 'new' old value
    */

    for (i = 0; i < recorders[mode]->rec_count; i++) 
	if (recorders[mode]->recs[i].entry == X)
	    return NULL;
    

    return record1(recorders[mode],X,tag);
}					    

void hist_record_add_line(rec,rest)
     struct hist_record *rec; 
     char *rest;
{
    rec->old_line = safe_realloc(rec->old_line,
				 sizeof(rec->old_line[0]) *
				 ( rec->line_count + 1));
    
    rec->old_line [ rec->line_count ] = safe_strdup(rest);
    rec->line_count++;
}

void hist_record_first_line(rec,rest)
     struct hist_record *rec; 
     char *rest;
{
    if (rec->line_count)
	panic("RC PANIC",__FILE__,__LINE__,"hist_record_first_line",
	      "Bad record",0);
    
    rec->old_line = safe_malloc(sizeof(rec->old_line[0]));
    
    rec->old_line[0] = safe_strdup(rest);
    
    rec->line_count = 1;
}

void hist_record_changed(rec)
     struct hist_record *rec;
{
    rec->changed = 1;
}


int seed_history(F,filename,recorder)
     FILE *F; 
     char * filename;
     struct elmrc_recorder  *recorder;
{
    int ok = 1;
    char * buffer = NULL;
    int r;  /* length of line or error/eof status */
    int lineno = 0;
    char current_tag[SLEN];

    struct rc_save_info_rec * prev_type = NULL;
    struct hist_record      * rec       = NULL;
    int ignored = 0;

    current_tag[0] = '\0';

    /* malloc_gets does realloc, so buffer need not
       be freed on loop 
    */
    while (0 <= ( r = malloc_gets(&buffer,MAX_LINE,F))) {
	char   * tag         = NULL;
	char   keyword[SLEN];
	char   * rest        = NULL;
	int    restlen = 0;
	int    assig         = 0;
	
	int    negate = 0;

	lineno++;

	if (!buffer)           /* EOF           */
	    break;

	if (buffer[0] == '#'   /* comment       */
	    || r < 1)          /* empty line    */
	    continue;

	rest = rc_split_keyword(buffer,r,keyword, sizeof keyword,
			     &tag,&assig, &restlen);


	if (tag && !keyword[0]) {
	    strfcpy(current_tag, tag, sizeof current_tag);

	    if (rest[0] && '#' != rest[0]) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmIgnoredLine,
				  "Rest of line ignored after \"%s:\" on line %d on file %s"),
			  tag, lineno,filename);
		ok = 0;
	    }

	} else if (!keyword[0]) {
	    if (prev_type != NULL) {
		
		if (RCTYPE_magic != prev_type->dt_type->magic)
		    panic("RC PANIC",__FILE__,__LINE__,"seed_history",
			  "Bad prev_type",0);
		
		if (rec) {
		    rec->old_line = safe_realloc(rec->old_line,
						 sizeof(rec->old_line[0]) *
						 ( rec->line_count + 1));
		    
		    rec->old_line [ rec->line_count ] = safe_strdup(rest);
		    rec->line_count++;

		    /* Preserve seeded data when printing again */
		    rec->changed++;
		}
	    }

	} else if (assig) {

	    if (tag || current_tag[0]) {
		struct rc_save_info_rec * rc_options      = NULL;
		int                     rc_option_count   = 0;
		
		if (!tag) 
		    tag = current_tag;
		else
		    ignored = 0;
		
#ifdef USE_DLOPEN
		if (give_options(tag,&rc_options,&rc_option_count)) {
		    
		    prev_type = rc_locate_option(rc_options,rc_option_count,
					      keyword,&negate);
		    ignored = 0;
		    
		    if (!prev_type) {
			lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadKeyTagInElmrc,
					  "I can't understand keyword \"%s\" (section %s) in line %d in \"%s\" file"),
				  keyword, tag,lineno,filename);
			ok = 0;
			continue;
		    }
		    
		} else if (!ignored) {
		    ignored = 1;
		    
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmIgnoredSection,
				      "Keywords on section \"%s\" ignored starting from line %d on file %s"),
			      tag, lineno,filename);
		    ok = 0;
		}
#else
		if (!ignored) {
		    ignored = 1;
		    
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSharedIgnoredSection,
				      "Shared libraries not supported. Section \"%s\" ignored starting from line %d on file %s"),
			      tag, lineno,filename);
		    ok = 0;
		}
#endif
		
	    } else {
		
		prev_type = rc_locate_option(save_info,NUMBER_OF_SAVEABLE_OPTIONS,
					  keyword,&negate);
		
		if (!prev_type) {	       
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadKeyInElmrc,
				      "I can't understand keyword \"%s\" in line %d in \"%s\" file"),
			      keyword, lineno,filename);
		    ok = 0;
		    continue;
		}
		
	    }
	    
	    rec = record1(recorder,prev_type,tag);
	    if (rec) {
		if (rec->line_count)
		    panic("RC PANIC",__FILE__,__LINE__,"seed_history",
			  "Bad prev_type",0);
		
		rec->old_line = safe_malloc(sizeof(rec->old_line[0]));
		
		rec->old_line[0] = safe_strdup(rest);
		
		rec->line_count = 1;
		rec->changed++;
	    }
	} else {
	    prev_type = NULL;

	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadLineInElmrc,
			      "Keyword \"%s\" without = -character in line %d in \"%s\" file"),
		      keyword, lineno,filename);
	    ok = 0;

	}
    }

    if (-1 == r) {	
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTooLongLine,
			  "%.30s: %d: Too long line: %.30s..."),
		  filename,lineno+1,
		  buffer ? buffer : "<not available>");
	ok = 0;
    }

    if (ferror(F)) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmErrorReading,
			  "Error reading from %.50s"),
		  filename);
	ok = 0;
    }

    if (buffer)
	free(buffer);


    return ok;
}

static void print_history_record P_((FILE *F,struct hist_record *x));
static void print_history_record(F,x)
     FILE *F;
     struct hist_record *x;
{
    int i;

    fprintf(F,"\n%s = %s\n",
	    x->entry->name,
	    x->old_line[0]);

    for (i = 1; i < x->line_count; i++) {
	fprintf(F,"\t%s\n",
		x->old_line[i]);
    }


}

void print_history_changed(F,recorder)
     FILE *F; 
     struct elmrc_recorder  *recorder;
{
    int i;

    /* First print lines without tag */
    char * last_tag = NULL;


    for (i = 0; i < recorder->rec_count; i++) {

	if (recorder->recs[i].changed &&
	    NULL == recorder->recs[i].tag) {
	    print_history_record(F,& (recorder->recs[i]));
	}

    }

    /* Then print values with tag */

    for (i = 0; i < recorder->rec_count; i++) {

	if (recorder->recs[i].changed &&
	    NULL != recorder->recs[i].tag) {

	    if (!last_tag || 
		0 != strcmp(last_tag,
			    recorder->recs[i].tag)) {
		last_tag = recorder->recs[i].tag;

		fprintf(F,"\n\n%s:\n",
			last_tag);

		print_history_record(F,& (recorder->recs[i]));

	    }
	}
    }
}

void mark_XX(x)
     struct rc_save_info_rec *x;
{
    enum record_mode mode; 

    int found = 0;

    x->flags |= FL_LOCAL;
    x->flags |= FL_CHANGED;

    DPRINT(Debug,10,(&Debug,
		     "mark_XX: Setting %p %s changed\n",
		     x,x->name));
	
    for (mode = SYSTEM_RC; mode < RC_MODE_COUNT; mode++) {
    
	if (recorders[mode]) {
	    int i;

	    for (i = 0; i < recorders[mode]->rec_count; i++) 
		if (recorders[mode]->recs[i].entry == x) {
		    recorders[mode]->recs[i].changed++;
		    found++;

		    DPRINT(Debug,10,(&Debug,
				     "mark_XX: recorders[%d]->recs[%d].entry == %p\n",
				     mode,i,x));

		}
	} else {
	    DPRINT(Debug,10,(&Debug,
			     "mark_XX: no mode=%d\n",mode));
	}
    }
	
    if (!found) {
	DPRINT(Debug,10,(&Debug,
			 "mark_XX: no change record\n"));
    }

}

void mark_local_changed(A)
     void *A;
{
    int x;
    

    for (x = 0; x < NUMBER_OF_SAVEABLE_OPTIONS; x++) {
        if (save_info[x].val.dummy == A) {

	    DPRINT(Debug,10,(&Debug,
			     "mark_local_changed: %d: %p\n",
			     x,A));

	    mark_XX(& (save_info[x]));

	}
    }

#ifdef USE_DLOPEN
    mark_shared_changed(A);
#endif

    mark_delayed_changed(A);

}




void mark_flocal_changed(A)
     option_func *A;
{
    int x;
    

    for (x = 0; x < NUMBER_OF_SAVEABLE_OPTIONS; x++) {
        if (save_info[x].func_val == A) {

	    DPRINT(Debug,10,(&Debug,
			     "mark_flocal_changed: %d: %p\n",
			     x,A));

	    mark_XX(& (save_info[x]));

	}
    }

#ifdef USE_DLOPEN
    mark_fshared_changed(A);
#endif

    mark_fdelayed_changed(A);
}



#ifdef HAVE_CODESET
#include <langinfo.h>
#endif

static int lineno = 0;
static int errors = 0;

static void do_expand_env P_((char *, char *, char *, unsigned));

static charset_t set_elmrc_charset P_((const char *filename,int *errors));

int parse_elmrc(F,global,filename)
     FILE *F;
     int global;
     char *filename;
{
    struct parsed_rc *rc = NULL;
    charset_t   rcset = NULL;

    errors = 0;

    elmrc_charset[0] = '\0';   /* Reset old value */

    rc = do_rc_parse(F, global ? SYSTEM_RC : LOCAL_RC, filename,&errors);

    rcset = set_elmrc_charset(filename,&errors);

    if (rc) {
	do_rc_process(rc, global ? SYSTEM_RC : LOCAL_RC,filename,
		      rcset,&errors);
	free_parsed_rc(&rc);
    } else
	errors++;	

    return errors;
}
#ifndef ok_rc_char
#define ok_rc_char(c)   (isalnum(c) || c == '-' || c == '_')
#endif

char old_system_text_file[SLEN]     = OLD_SYSTEM_TEXT_FILE;
char old_system_data_file[SLEN]     = OLD_SYSTEM_DATA_FILE;
char system_aliases_file[SLEN]      = SYSTEM_ALIASES_FILE;
char system_rc_file[SLEN]       = SYSTEM_RC_FILE;
char system_mime_types[SLEN]    = SYSTEM_MIME_TYPES;
char system_iso2022_sets[SLEN]  = SYSTEM_ISO2022_SETS;
char system_terminal_info[SLEN] = SYSTEM_TERMINAL_INFO;
char system_mail_services[SLEN] = SYSTEM_MAIL_SERVICES;
char system_mlist_file[SLEN]    = SYSTEM_MLIST_FILE;
char hostdomfile[SLEN]          = HOSTDOMFILE;
char system_mime_charsets[SLEN] = SYSTEM_MIME_CHARSETS;

ESTR map_txtdir_e = {            /* init_defaults */
    ESTR_meta|ESTR_slash,
    NULL,
    NULL
};
ESTR map_bindir_e = {            /* init_defaults */
    ESTR_meta|ESTR_slash,
    NULL,
    NULL
};

char unidata_path[SLEN];        /* init_defaults */
char unidata_path_global[SLEN];   /* Used before user elm.rc is read */
char raw_unidata_path[SLEN] =  "unidata.bin";     /* init_defaults */

char bindata_path_global[SLEN];   /* Used before user elm.rc is read */
char bindata_path[SLEN];  /* init_defaults */
char raw_bindata_path[SLEN] =  "bindata.bin";     /* init_defaults */

int  unstable_reverse_thread = 0;

char user_rc_file[SLEN]   = ELMRCFILE;              /* user_init fixes */
char old_user_text_file[SLEN] = OLD_USER_ALIAS_TEXT;     /* user_init fixes */
char old_user_data_file[SLEN] = OLD_USER_ALIAS_DATA;     /* user_init fixes */
char user_aliases_file[SLEN] = USER_ALIASES_FILE;        /* user_init fixes */
char user_mime_types[SLEN] = USER_MIME_TYPES;       /* user_init fixes */
char user_mailheaders[SLEN] = MAILHEADERS;          /* user_init fixes */
char user_mime_charsets[SLEN] = USER_MIME_CHARSETS; /* user_init fixes */
char user_iso2022_sets[SLEN]  = USER_ISO2022_SETS;  /* user_init fixes */
char user_terminal_info[SLEN] = USER_TERMINAL_INFO; /* user_init fixes */
char user_mlist_file[SLEN]    = USER_MLIST_FILE;    /* user_init fixes */
char user_mail_services[SLEN] = USER_MAIL_SERVICES; /* user_init fixes */

ESTR defaultfile_e; 	                /* name of default folder */
ESTR calendar_file_e; 	                /* name of file for clndr  */
ESTR dead_letter_dir_e; 
int clear_pages ;		/* flag: act like "page" (more -c)? */
int title_messages = 1;		/* flag: title message display?     */

nl_catd  elm_msg_cat; 	 /* message catalog	    */
                         /* WARNING: Elm's version uses 0 as failure
                                     System uses (nl_catd) -1 as failure !
			 */


char username[SLEN];	/* return address name!    */
char home[SLEN];		/* home directory of user  */
int userid = -1;			/* uid for current user	      */
int groupid = -1;			/* groupid for current user   */

int mailgroupid;		/* groupid for current user   */
int have_saved_ids;             /* sysconf _SC_SAVED_IDS      */

/* ------------- Variables in elmrc ------------------------------- */

int  add_irt_phrase;          /* In-reply-to: Add phare to in-reply-to ?
			           True = Do not follow RFC 2822 
				*/
int add_sender_header = 1;     /* If set add Sender: header if user specifies
				  From: -header
				*/
int allow_charset_switching;   /* flag: allow changing charset of */
				 /* terminal if terminal supports   */
int allow_setuid;

static struct sortval ASR_srtval[]={
    {"alias", ALIAS_SORT},
    {"name", NAME_SORT},
    {"text", TEXT_SORT},
    {NULL, 0} 
};

SORT alias_sortby = {            /* how to sort aliases...   */
    NAME_SORT,	           

    ASR_srtval
};

int alias_group_phrase = 1;      /* Use phrases on group aliases */ 

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

static  struct keyword_val ALTEDITOR_KW[] = {
    { "EDITOR",       alteditor_kw_EDITOR, 0 },
    { "use-$EDITOR",  alteditor_KW_EDITOR, 1 },    /* bitmask */
    { "use-$VISUAL",  alteditor_KW_VISUAL, 1 },    /* bitmask */
    { NULL,           alteditor_kw_NO }
};
    
/* Returns 1 if success, 0 failure, -1 if expansion is not string */
S_(keyword_expand_func alteditor_expand_func)
static int alteditor_expand_func P_((struct dt_estr_info *kw,
				     struct dt_keyword_info *ki,
				     unsigned int keyword,
				     unsigned int continue_mask)); 
static struct dt_keyword_info alteditor_keyword_info = {
    alteditor_expand_func,
    ALTEDITOR_KW
};

ESTR alternative_editor_e = {    /* alternative editor...    */
    ESTR_none|ESTR_print_expanded,
    NULL,NULL,
    alteditor_kw_EDITOR, alteditor_KW_EDITOR|alteditor_KW_VISUAL,
    & alteditor_keyword_info,
    0
};

PATH alternative_addresses = {	/* how else do we get mail? */
    PATH_sep_comma| PATH_ascii_only,
    NULL, 0,NULL
};

int always_del;		/* flag: always delete marked msgs? */
int always_keep = 1;		/* flag: always keep unread msgs?   */
int always_store;		/* flag: always store read msgs?    */
int ask_delete = 1;             /* confirm delete on resync? */
int ask_keep = 1;              /* ask to keep unread msgs? */
int ask_reply_copy = 1;        /* ask to copy mssg into reply? */
int ask_store = 1;             /* ask to store read mail? */
int arrow_cursor;		/* flag: use "->" cursor regardless?*/
int prompt_for_cc = 1;		/* flag: ask user for "cc:" value?  */

#ifdef USE_PGP
int pgp_askpgpsig; /* Should pgp ask userid to sign messages with? */

static char * PGP_SIGN_TYPES[] = { "application/pgp", "text/plain", 
				   "text/x-pgp", NULL };


ENUMERATE pgp_sign_type= {     /* 0 = application/pgp  */
    pgp_sign_application_pgp,  /* 1 = text/plain       */
    NUM_pgp_sign_type,         /* 2 = text/x-pgp       */ 
    &(PGP_SIGN_TYPES[0])    
};

static char * PGP_ENCRYPT_TYPES[] = { "application/pgp", "text/plain", 
				      "text/x-pgp", NULL };

ENUMERATE pgp_encrypt_type= {   /* 0 = application/pgp */
    pgp_encrypt_application_pgp,/* 1 = text/plain      */
    NUM_pgp_encrypt_type,       /* 2 = text/x-pgp      */
    &(PGP_ENCRYPT_TYPES[0])  
};

int pgp_interactive;            /* Should pgp/gpg encryption/signing to be
			            run always on interactive mode */

static char * PGP_VERSIONS[] = { "default", "pgp2", "pgp5", "gpg", NULL };
ENUMERATE send_pgp_version = {  /* preferred pgp version*/
    pgp_none,
    PGP_NUM, &(PGP_VERSIONS[0]) 
};
	
#endif

struct string * attribution_s;	        /* attribution string for replies   */
struct string * fwdattribution_s;	/* attribution string for forwarded 
					   mssgs */

static char * AUTO_ATTACH[] = { "none", "application", "non-text", NULL };
ENUMERATE  auto_attachment = {   /* 0 = none,                     */
    auto_attachment_application, /* 1 = application,              */
    NUM_auto_attachment,         /* 2 = non-text   		  */
    &(AUTO_ATTACH[0])	 
};

int auto_copy;		        /* flag: automatically copy source? */
int reply_copy = 1;		/* flag: copy message into reply */

int auto_iso_8859 = 1;          /* flag: Should ISO-8859-X charsets defined
				   on demand ?                      */
#ifdef BACKGROUD_PROCESSES       
int background_wait_time = 2;    /* If > 0 background mailer after  */
				 /* this number of seconds          */
#endif

static char * BROWSER_WILDCARD[] = { "off", "on", "file browser" };
ENUMERATE browser_wildcards = {  /* do wildcard match?         */
    browser_wildcard_yes,
    NUM_browser_wildcards, &(BROWSER_WILDCARD[0])
};

int browser_tabcomplete = 1;    /* flag:  tabcomplete prompt?       */
int builtin_lines= -3;		/* int: if < 0 use builtin if message*/
                                /* shorter than LINES+builtin_lines */
                                /* else use pager. If > 0 use builtin*/
                                /* if message has fewer than # of lines*/
int charset_convert_ok;         /* flag: no warning about conversion? */
int check_type_pattern = 1;     /* flag: check magic numbers when paging */
char raw_text_charset[SLEN] = "DISPLAY";    /* name of character set     */
charset_t text_charset;   	    /* pointer of character set  */
char raw_default_mimetext_charset[SLEN] = "US-ASCII"; 
                                                /* name of character set    */
charset_t default_mimetext_charset;      /* pointer to character set */
char raw_default_nomime_charset[SLEN] = "SYSTEM"; /* name of character set  */
charset_t default_nomime_charset;      /* pointer to character set   */

static char * FOLDER_STATUS[] = { "Read", "Old", "New", NULL };
ENUMERATE def_folder_status = {      /* 0 = Read                    */
    FOLDER_STATUS_OLD,               /* 1 = Old                     */
    NUM_folder_status,               /*	2 = New                     */
    &(FOLDER_STATUS[0])           
};

char config_options[SLEN];/* which options are in o)ptions    */

int confirm_append;		/* flag: confirm append to folder?  */
int confirm_create;		/* flag: confirm create new folder? */
int confirm_files;		/* flag: confirm files for append?  */
int confirm_folders;	        /* flag: confirm folders for create?*/
int confirm_print;              /* flag: confirm printing of messages ?    */
int confirm_tag_save = 1;       /* flag: confirm saving of tagged messages */
int edit_flush = 1;	        /* flag: flush input after extern edit? */
int convert_comment = 1;        /* flag: convert comment to fullname ?  */
int convert_utf_header = 1; /* flag: convert utf-8 mime encoded words to utf-7 when sending?   */

int auto_copy_sent = -1;	  /* flag: automatically copy sent mail? */
char display_locale[SLEN] = "NONE";  /* LC_CTYPE locale (character set) */
char display_codeset[SLEN] = "";     /* nl_langinfo(CODESET) for current locale (character set) */

charset_t wanted_display_charset;
char raw_display_charset[SLEN]   = "SYSTEM";	
                               /* the wanted display charset set */

int DSN_success;               /* flag: Ask successfull DSNes      */
char e_editor[SLEN];	/* "~e" editor...                   */

static  struct keyword_val EDITOR_KW[] = {
    { "builtin",    editor_kw_builtin,   0 },
    { "internal",   editor_kw_builtin,   0 },
    { "use-$EDITOR", editor_KW_EDITOR,  1 },    /* bitmask */
    { "use-$VISUAL", editor_KW_VISUAL,  1 },    /* bitmask */
    { NULL,          editor_kw_NO }
};

/* Returns 1 if success, 0 failure, -1 if expansion is not string */
S_(keyword_expand_func editor_expand_func)
static int editor_expand_func P_((struct dt_estr_info *kw,
				  struct dt_keyword_info *ki,
				  unsigned int keyword,
				  unsigned int continue_mask));
static struct dt_keyword_info editor_keyword_info = {
    editor_expand_func,
    EDITOR_KW
};


ESTR editor_e = {                 /* editor for outgoing mail         */
    ESTR_none|ESTR_print_expanded,
    NULL,NULL,
    0, alteditor_KW_EDITOR|alteditor_KW_VISUAL,
    & editor_keyword_info,
    0
};

static char * EDITOR_PROPLINE [] = 
    { "auto", "emacs", "ignore", NULL };
ENUMERATE editor_elmrc_propline = {
    editor_propline_auto,
    NUM_editor_propline,
    &(EDITOR_PROPLINE[0])
};


static char * MPAR[] = { "plain", "encoded", "plain-and-encoded" };
ENUMERATE mime_parameters = {
    mime_parameters_plain_and_encoded,
    NUM_mime_parameters, &(MPAR[0])
};

static char * ENV_FROM_SOURCES[] = { "forward-from", "from", "return-path", 
				     NULL };
ENUMERATE env_from_source = {  /* 0 == forward-from,                */
    env_source_forward_from,   /* 1 == from,                        */
    NUM_env_from_source,       /* 2 == return-path                  */
    &(ENV_FROM_SOURCES[0])  
};

char escape_char = TILDE_ESCAPE;/* '~' or something else..          */
int force_name;		     /* flag: save by name forced?	    */

static char * FRAGMENTH_list[] = { "none", "manual", "auto" };
ENUMERATE fragment_handling = { /* 0 == none,                       */
    fragment_handling_manual,   /* 1 == manual,                     */
    NUM_fragment_handling,      /* 2 == auto                        */
    &(FRAGMENTH_list[0])     
};
struct string * full_username_s; /* Full username - gecos           */

static char * HDR_PHRASE_DISPLAY_MODES[] = { "plain", "quoted", NULL };
ENUMERATE phrase_display_mode = {    /* 0 == plain, 1 = quoted */
    phrase_display_plain,
    NUM_phrase_display_mode, &(HDR_PHRASE_DISPLAY_MODES[0]) 
};

char hostdomain[SLEN];	/* name of domain we're in          */
char hostfullname[SLEN];	/* name of FQDN we're in            */
char hostname[SLEN];	/* name of machine we're on         */
#ifdef DISP_HOST
int  menu_display_host = TRUE;  /* flag: Display host on mailbox menu* */
#else
int  menu_display_host = FALSE;  /* flag: Display host on mailbox menu? */
#endif
int keep_empty_files;	        /* flag: leave empty folder files?  */
#ifdef USE_PGP
int pgp_keeppassfor = 300;      /* 5 minutes                        */
#endif

static struct sortval DIR_srtval[]={
    {"none", LD_NONE},
    {"name", LD_NAME_SORT},
    {"mtime", LD_MTIME_SORT},
    {NULL, 0} 
};

SORT local_dir_sortby = {            /* how to sort files...  */
    LD_NAME_SORT,	           
    DIR_srtval
};

int local_fast_lookup;          /* flag: directory listing not needed 
				 *       for lookup 
				 */

int fb_show_dotfiles;           /* flag: Show .dotfiles on Elm 2.5 
                                 *       style browser
				 */

int fb_show_protection = 1;     /* flag: Show permission and owner columns 
				 *  on Elm 2.5 style browser 
				 */

int fb_show_title = -1;         /* flag: Show title of columns on Elm 2.5 
				 *       style browser 
				 */

#define DEFAULT_SORT_ORDER    SHRT_MAX
SORT fb_sort_order = {            /* how to sort files...  */
    DEFAULT_SORT_ORDER,	           
    DIR_srtval
};

#ifdef PIDCHECK
static char * LLP_list[] = { "no", "yes", "hostname", 
			     "ignore-hostname", "require-hostname", NULL };
ENUMERATE local_lockfile_pidcheck = {     /* 0 == no               */
    pidcheck_hostname,			  /* 1 == yes              */
    NUM_lockfile_pidcheck, &(LLP_list[0]) /* 2 == hostname         */
};					  /* 3 == ignore-hostname  */
                                          /* 4 == require-hostname */
#endif

char local_sessionlock_dir[SLEN] = default_temp;

static char * LSL_list[] = { "never", "spool", "always", "non-spool", NULL };
ENUMERATE local_sessionlock_use_home = {  /* 0 == never,          */
    local_seslck_use_home_never,          /* 1 == spool           */
    NUM_local_seslck_use_home,            /* 2 == always          */ 
    &(LSL_list[0])			  /* 3 == non-spool  	  */
};					  


char raw_local_fs_charset[SLEN] = "SYSTEM"; /* filesystem charset */
charset_t local_fs_charset;     /* filesystem charset */
ESTR local_signature_e;         /* local msg signature file     */

#if defined(WCHAR) && defined(__STDC_ISO_10646__)
PATH locale_charsets;           /* Use type=unknown for these charsets
			         * if system charset
			         */
#endif

#ifdef SYSCALL_LOCKING
int lockfolders  = 1;           /* Lock folder when open */
int lock_in_copy = 1;           /* Lock folder when copied to it    */
#else
int lockfolders;                /* Lock folder when open */
int lock_in_copy;               /* Lock folder when copied to it    */
#endif

int long_encoded_headers = 1;   /* Produce over 80 character mime encoded
				   headers */

ESTR folders_e = 	{       /* folder home directory            */
    ESTR_slash,
    NULL,
    NULL
};   

ESTR extra_mailbox_dir_e = {
    ESTR_slash,
    NULL,
    NULL
};
ESTR attachment_dir_e = {   /* {doc} -prefix */
    0,                      /* Note that default assigment do not add / to end */
    NULL,
    NULL
};

int attachment_fix_extension = 1; /* flag: fix extension of attashment when saving */
PATH mailname = {                 /* Name used after the @  -character on addresses and 
				     aliases of that. */
    PATH_sep_comma| PATH_ascii_only,
    NULL, 0,NULL
};

int mail_permissions = 0600;	  /* int: permissions for mailbox files */
int mini_menu_rc             = 1; /* flag: menu specified?	    */

static char * hide_hacks[] = { "none", "FOLDER INTERNAL DATA", NULL };
ENUMERATE message_hide_hack = { /* 0 == none                          */
    message_hide_none,          /* 1 = FOLDER INTERNAL DATA           */
    NUM_message_hide_hack, &(hide_hacks[0])
};

ESTR metamail_path_e = {        /* Metamail path or "none"          */
     ESTR_none|ESTR_metamail,   /* if no metamail                   */
     NULL,
     NULL
};
PATH metamail_mailcaps = {   /* mailcaps for metamail */
    PATH_file|PATH_sep_colon|PATH_expand,
    NULL, 0,NULL
};

PATH internal_mailcaps = {   /* mailcaps for internal handler */
    PATH_file|PATH_sep_colon|PATH_expand,
    NULL, 0,NULL
};

PATH internal_mailcap_t_programs;
int internal_mailcap_t_prompt;

PATH incomingfolders = {      /* treate as mailbox */
    PATH_file|PATH_expand|PATH_sep_comma,
    NULL, 0, NULL
};

PATH mailcap_bl_programs;       /* Do not execute program or call metamail */

int metoo;			/* flag: copy me on mail to alias?  */
int mime_body_keywords = TRUE;  /* flag: parse body looking for encode */
				/* keywords */
int mimeforward = FALSE;        /* flag: Forward as Message/rfc822  */
int askmimeforward = 1;         /* flag: prompt about messaga/rfc822 forward */
int move_when_paged;	        /* flag: move when '+' or '-' used? */
int names_only;		        /* flag: display user names only?   */

static char * NOENCODINGS[] = { "pass-7bit", "pass-8bit", "pass-binary", 
				NULL };
ENUMERATE allow_no_encoding = { /* 1: Allow 8bit without -B8BITMIME */
    allow_pass_7bit,            /* 2: Allow binary without -BBINARYMIME and */
    NUM_allow_no_encoding,      /*    and 8bit without -B8BITMIME   */
    &(NOENCODINGS[0])        
};

int allow_no_hdrencoding;       /* TRUE, if header encoding is      */
				/* not required                     */
int noheader = 1;		/* flag: copy + header to file?     */
int noheaderfwd;		/* flag: copy + header to file(fwd)? */
int pagealternative=1;          /* Page if alternative have one know*/
                                /* subtype ?                        */
int page_known_charsets=1;      /* Page if charset is not unknown?  */
int pagemultipart;               /* Page unknown mime subparts?      */
int pagesigned;                  /* Page unknown signed protocols?   */

static  struct keyword_val PAGER_KW[] = {
    { "builtin",    pager_kw_builtin,   0 },
    { "builtin+",   pager_kw_builtin,   0 },
    { "builtin++",  pager_kw_builtin,   0 },
    { "internal",   pager_kw_builtin,   0 },
    { "use-$PAGER", pager_KW_PAGER,  1 },    /* bitmask */
    { NULL, pager_kw_NO }
};

/* Returns 1 if success, 0 failure, -1 if expansion is not string */
S_(keyword_expand_func pager_expand_func)
static int pager_expand_func P_((struct dt_estr_info *kw,
				 struct dt_keyword_info *ki,
				 unsigned int keyword,
				 unsigned int continue_mask));


static struct dt_keyword_info pager_keyword_info = {
    pager_expand_func,
    PAGER_KW
};


ESTR pager_e = {                 /* what pager to use                */
    ESTR_none|ESTR_print_expanded,
    NULL,
    NULL,

#ifdef USE_BUILTIN_PAGER  
    pager_kw_builtin,     0,
#else
    0,                    pager_KW_PAGER,
#endif
    & pager_keyword_info,
    0
};

int pager_indicate_wrapping = 1; /* Idicate wrapping with \ on builtin
					pager                               */

int pager_paragraph_width = 70;  /* Indicate preferred width for flowed 
				    (word wrapped) paragraph. */

#ifdef USE_PGP
ESTR pgp2_path_e = {            /* Pgp path or "none" if no pgp     */
    ESTR_none,
    NULL,
    NULL
};
ESTR pgp5_dir_e = {   /* pgp5 binary directory or "none" if no pgp5 */
    ESTR_slash|ESTR_none,
    NULL,
    NULL    
};
ESTR gpg_path_e  = {
    ESTR_none,
    NULL,
    NULL
};
#endif
int point_to_new = 1;		/* flag: start pointing at new msg? */
char allowed_precedences[SLEN];	/* list of precedences user */
                                /* may specify                      */
ESTR printout_e = {	        /* how to print messages            */
    ESTR_none,
    NULL,
    NULL
};
char printhdrs[SLEN];
int prompt_after_pager = 1;     /* flag: prompt after pager exits   */
int prompt_after_metamail = 1;  /* flag: prompt after metamail exits     */
int  prompt_metamail = 1;     /* flag: prompt for calling of metamil */
int quote_forward;		/* flag: fwd'd msgs quoted like replies */
struct string * quote_prefix;     /* prefix char(s) for msgs          */
int readdatapercentinc = 5;	/* data percent increment during new mbox read */

ESTR readmsg_e = {	        /* path to readmsg program	*/
    ESTR_meta,
    NULL,
    NULL
};

int readmsginc = 100;		/* increment of msg cnt when reading*/
                                /* new mbox                         */
ESTR recvd_mail_e = {	/* folder for storing received mail */
    ESTR_none|ESTR_meta,
    NULL,
    NULL
};

ESTR remote_signature_e;        /* remote msg signature file    */
int req_mime_bodyencoding = 1;  /*                                  */
int req_mime_hdrencoding;       /*                                  */
int resolve_mode = 1;		/* flag: delete saved mail?	    */
int save_by_name = 1;		/* flag: save mail by login name?   */
int save_by_alias = 1;	/* save mail by alias of login name? */

ESTR sent_mail_e  = {	        /* name of file to save copies to   */
    ESTR_none|ESTR_meta,
    NULL,
    NULL
};

int send_mime_plain  = 1;       /* flag: send plain text as MIME?   */

ESTR shell_e  = {	        /* current system shell             */
    ESTR_bin,
    NULL,
    NULL
};

int  set_window_title;          /* flag: set (xterm?) window title and 
				   icon name */
#ifdef USE_PGP
int pgp_noarmor=1;      /* Should Elm display text before PGP armor */
#endif
int showto;                     /*                                  */
int show_reply = 1;	        /* flag: show 'r' for replied mail  */
int show_mlists;	        /* show mailing list info?          */

static char * HEADER_ERROR_MODES[] = 
    { "OFF", "ON", "store", NULL };
ENUMERATE show_header_errors_e = {
    show_hdr_error_store,
    NUM_show_header_errors, &(HEADER_ERROR_MODES[0]),
};

int sig_dashes = 1;		/* flag: include dashes above sigs? */
int sleepmsg = 2;		/* time to sleep for messages being */
                                /* overwritten on screen            */

static struct sortval SRT_srtval[]={
    {"sent", SENT_DATE},
    {"thread", THREAD},
    {"received", RECEIVED_DATE},
    {"recieved", RECEIVED_DATE},
    {"rec", RECEIVED_DATE},
    {"from", SENDER},
    {"sender", SENDER},
    {"size", SIZE},
    {"lines", SIZE},
    {"subject", SUBJECT},
    {"mailbox", MAILBOX_ORDER},
    {"folder", MAILBOX_ORDER},
    {"status", STATUS},
    {NULL, 0} 
};

SORT sortby = {                 /* how to sort incoming mail...     */
    REVERSE SENT_DATE,	

    SRT_srtval
};

static struct sortval THREAD_srtval[]={
    {"sent",    SENT_DATE},
    {"subject", SUBJECT},
    {NULL, 0} 
};


SORT thread_sortby = {          /* how to sort threads of messages */
    REVERSE SENT_DATE,	

    THREAD_srtval
} ;      

long elm_timeout = 600L;	/* timeout (secs) on main prompt    */
int sort_thread_max_time = 30;  /* Number of days which can considered 
				    be same thread */
char to_chars[SLEN] = DEFAULT_TO_CHARS;	/* chars to indicate who mail is to */

ESTR temp_dir_e  = {         /* name of temp directory           */
    ESTR_slash,
    NULL,
    NULL
};

#ifdef USE_PGP
int pgp_keeppass;             /* should Elm keep the passphrase in*/
                                /* memory?                          */
#endif

static char * USERLEVELS[] = { "beginner", "intermediate", "expert", NULL };
ENUMERATE user_level = { 	/* flag: how good is the user?      */
    user_level_beginner,
    NUM_user_level, &(USERLEVELS[0]) 
};

int use_tite = 1;		/* flag: use termcap/terminfo ti/te?*/
int utf7_encode_optional = 1;   /* flag: Should utf7 optional direct 
				   characters to be encoded?        */
#ifdef I_NETDB
int verify_domain;              /* flag: Should we try check existence
				   of mail domain
				*/
#endif
int verify_local_address;       /* flag: Should we try check existence
				    of local user
				 */

char v_editor[SLEN];	        /* "~v" editor...                   */
int elm_filter = 1;		/* flag: weed out header lines?	    */
PATH weedlist = {
    PATH_sep_comma|PATH_quote|PATH_convert_underline,
    NULL, 0,NULL
};

#ifdef REMOTE_MBX
static struct sortval IMAP_srtval[]={
    {"none", ID_NONE},
    {"name", ID_NAME_SORT},
    {NULL, 0} 
};

SORT imap_dir_sortby = {            /* how to sort imap folders...  */
    ID_NONE,	           
    IMAP_srtval
};

char raw_imap_charset[SLEN] = "SYSTEM";  /* IMAP foldername charset */
int imap_fast_lookup;                /* flag: skip some directory listing */
int imap_max_dl_size = 4*1024;      /* Maximum size of message to be
				     * downloaded 
				     */
charset_t imap_charset = NULL;           /* IMAP foldername charset */
int IMAP_connection_cache = 1; /* flag: call cache_connection() ? */
int IMAP_name_convention = 1; /* flag: Should imap intearnational folder
				 convention be used? */
int IMAP_use_examine   = 1;  /* flag: Should EXAMINE command to be used?    */
int IMAP_show_greeting = 1;  /* flag: Should untagged OK messages be shown? */
int IMAP_show_warning  = 1;  /* flag: Should untagged NO messages be shown? */
int IMAP_show_error    = 1;  /* flag: Should untagged BAD messages be shown? */
int pop_max_dl_size = 4*1024;      /* Maximum size of message to be
				     * downloaded 
				     */
int POP_show_greeting  = 1;  /* flag: Should POP server greeting be shown?  */
#endif

/* ---------------------------------------------------------------- */

static void do_expand_env P_((char  *descr, char *dest, char *src,
			      unsigned	destlen));
static void do_expand_env(descr, dest, src, destlen)
     char	*descr, *dest, *src;
     unsigned	destlen;
{
    if (expand_env(dest, src, destlen) != 0) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCannotInitErrorExpanding,
			  "Cannot initialize \"%s\" - error expanding \"%s\"."),
		  descr, src);
	errors++;
	dest[0] = '\0';
    }
}

static void do_expand_meta P_((char  *descr, char *dest, char *src,
			      unsigned	destlen));
static void do_expand_meta(descr, dest, src, destlen)
    char	*descr, *dest, *src;
    unsigned	destlen;
{
    if (expand_meta(dest, src, destlen) != 0) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCannotInitErrorExpanding,
			  "Cannot initialize \"%s\" - error expanding \"%s\"."),
		  descr, src);
	errors++;
	dest[0] = '\0';
    }
}

static void expand_map_bin P_((char  *descr, char *dest, char *src,
			      unsigned	destlen));
static void expand_map_bin(descr, dest, src, destlen)
    char	*descr, *dest, *src;
    unsigned	destlen;
{
    if (strcmp(src,"none") == 0 || src[0] == '\0') {
	
	strfcpy(dest,"none",destlen);
	
    } else if (NULL == strpbrk(src,"/$\\;{}()")) {

	/* give_dt_estr_as_str adds / to end */
	const char * map_bindir_val = give_dt_estr_as_str(&map_bindir_e,
							  "map-bin-dir",NULL,NULL);

	if (!map_bindir_val) {
	    dest[0] = '\0';
	    return;
	}

	if (0 != access(map_bindir_val,ACCESS_EXISTS)) {
	    int err = errno;
	    lib_error(CATGETS(elm_msg_cat, ElmSet,ElmNoBinMapDirX,
			      "%s: no map-bin-dir exist: %s: %s"),
		      descr,
		      map_bindir_val,strerror(err));
	    
	    /* No error counter incrementging because 
	       map-bin-dir can be modified only on system 
	       elm.rc
	    */
	}
	elm_sfprintf(dest,destlen,
		     FRM("%s/%s"),
		     map_bindir_val,src);
	
    } else
	do_expand_meta(descr, dest, src, destlen);


    DPRINT(Debug,10,(&Debug, "%s expanded: %s => %s (destlen %d)\n",
		     descr,src,dest,destlen));


}

/* sets locale and elm_msg_cat */
static int locale_init_called = 0;
static int locale_error = 0;


void locale_init () {
#ifdef I_LOCALE
    char *res;
#endif
    
    DPRINT(Debug,9,(&Debug, 
		    "locale_init: called = %d\n",
		    locale_init_called));
    
#ifdef I_LOCALE

    if (setlocale(LC_ALL, "") == NULL) {
	lib_error(FRM("Elm: Unsupported locale (check $LANG)\n"));
	if (setlocale(LC_CTYPE,"") != NULL) {
	    lib_error(FRM("Elm: ... but succeed setting of LC_CTYPE\n"));
	} else {
	    lib_error(FRM("Elm: ... check also $LC_CTYPE\n"));
	}
	locale_error = 1;
    }
#endif
  
    /* WARNING: Elm's version uses 0 as failure
     *  System uses (nl_catd) -1 as failure !
     */


    elm_msg_cat       = catopen("elmme+2.5", 0);
    if (0            == elm_msg_cat ||
	(nl_catd) -1 == elm_msg_cat) {
	DPRINT(Debug,9,(&Debug, 
			"locale_init: catopen elmme+2.5 failed\n"));
    }
    /* NOTE:    elm_msg_cat    'leaks' on end of program   ... */

#ifdef  I_LOCALE
    res = setlocale(LC_CTYPE,NULL);
    if (!res) {
	lib_error(FRM("Elm: Unable to get current locale (LC_CTYPE)\n"));
	locale_error = 1;
    } else
	strfcpy(display_locale,res, sizeof display_locale);
#endif
  
#ifdef HAVE_CODESET
    res = nl_langinfo(CODESET);

  if (!res) {
      lib_error(FRM("Elm: Unable to get CODESET for current locale\n"));
      locale_error = 1;
  } else	  
      strfcpy(display_codeset,res, sizeof display_codeset);
#endif

  locale_init_called++;
}

enum matches { match_exact = 0, 
	       match_codeset,
	       match_charset, 
	       match_lang_country, match_lang, MATCH_COUNT };

struct charset_map_item  * system_charset_map = NULL;
struct charset_map_item  * user_charset_map   = NULL;

static void map_system_charset P_((const char * lc_ctype, 
				   const char * codeset,
				   int global_only));
static void map_system_charset(lc_ctype, codeset,global_only)
     const char * lc_ctype;
     const char * codeset;
     int global_only;
{  
    const char *lang = NULL;
    const char *lang_country = NULL;
    const char *charset = NULL;
    charset_t d_matches[MATCH_COUNT];
    int i,j;

    char * default_system_charset = DEFAULT_SYSTEM_CHARSET;
    char * c1;

    char lbuf[3], cbuf[6];

    DPRINT(Debug,3,(&Debug,
		    "map_display_charset: global_only=%d\n",global_only));

    if (!getenv("NOMETAMAIL") && (c1 = getenv("MM_CHARSET")) &&
	c1[0]) {

	default_system_charset  = c1;
	DPRINT(Debug,9,(&Debug, 
			"map_system_charset: using $MM_CHARSET (%s) as default system charset\n",
			default_system_charset));
    }

    for (i = 0; i < MATCH_COUNT; i++)
	d_matches[i] = NULL;

    if (isascii(lc_ctype[0]) && isalpha(lc_ctype[0]) &&
	isascii(lc_ctype[1]) && isalpha(lc_ctype[1]) &&
	('.' == lc_ctype[2] || '_' == lc_ctype[2] || '\0' == lc_ctype[2])) {
	strncpy(lbuf,lc_ctype,3);
	lbuf[2] = '\0';
	lang = lbuf;
	
	if ('_' == lc_ctype[2]) {
	    if (isascii(lc_ctype[3]) && isalpha(lc_ctype[3]) &&
		isascii(lc_ctype[4]) && isalpha(lc_ctype[4]) &&
		('.' == lc_ctype[5] || '\0' == lc_ctype[5])) {

		strncpy(cbuf,lc_ctype,6);
		cbuf[5] = '\0';
		lang_country = cbuf;
		
		if ('.' == lc_ctype[5])
		    charset = lc_ctype+6;
	    }
	} else if ('.' == lc_ctype[2])
	    charset = lc_ctype+3;      
    }
    
    if (codeset[0]) {
	DPRINT(Debug,10,(&Debug,
			 "map_system_charset: codeset: %s\n", 
			 codeset));
	d_matches[match_codeset] = codeset_name_to_charset(codeset);
	if (!d_matches[match_codeset]) {
	    DPRINT(Debug,1,(&Debug,
			    "map_system_charset: codeset name %s unknown!\n",codeset)); 
	}
    }

    if (lang) {
	DPRINT(Debug,10,(&Debug,
			 "map_system_charset: %s -> lang: %s\n", 
			 lc_ctype,lang));
    } 
    if (lang_country) {
	DPRINT(Debug,10,(&Debug,
			 "map_system_charset: %s -> lang_country: %s\n", 
			 lc_ctype,lang_country));
    }
    if (charset) {
	DPRINT(Debug,10,(&Debug,
			 "map_system_charset: %s -> charset: %s\n", 
			 lc_ctype,charset));
    }

    for (i = global_only ? 1 : 0; i < 3; i++) {
	if ( i < 2) {
	    int k;
	    struct charset_map_item * F = user_charset_map;
	    if (1 == i)
		F = system_charset_map;

	    if (!F)
		continue;

	    for (k = 0; F[k].charset; k++) {
		if (F[k].match) {
		    if (istrcmp (F[k].match, lc_ctype) == 0) {
			d_matches[match_exact] = F[k].charset;
			
			DPRINT(Debug,3,(&Debug,
					"map_system_charset: user defined \"%s\" as %s\n", 
					lc_ctype, 
					d_matches[match_exact]->MIME_name ?
					d_matches[match_exact]->MIME_name :
					"<no MIME name>"));
		    }
		    if (charset && strncmp(F[k].match,"*.",2) == 0  &&
			istrcmp (F[k].match+2, charset) == 0) {
			
			d_matches[match_charset] = F[k].charset;
			
			DPRINT(Debug,3,(&Debug,
					"map_system_charset: user defined \"%s\" (*.%s) as %s\n", 
					lc_ctype, charset,
					d_matches[match_charset]->MIME_name ?
					d_matches[match_charset]->MIME_name :
					"<no MIME name>"));	      
		    }
		    if (lang_country && istrcmp (F[k].match, lang_country) == 0) {
			d_matches[match_lang_country] = F[k].charset;

			DPRINT(Debug,3,(&Debug,
					"map_system_charset: user defined \"%s\" (%s) as %s\n", 
					lc_ctype, lang_country, 
					d_matches[match_lang_country]->MIME_name ?
					d_matches[match_lang_country]->MIME_name :
					"<no MIME name>")); 
			
		    }
		    
		    if (lang_country && istrcmp (F[k].match, lang) == 0) {
			d_matches[match_lang] = F[k].charset;
			
			DPRINT(Debug,3,(&Debug,
					"map_system_charset: user defined \"%s\" (%s) as %s\n", 
					lc_ctype, lang, 
					d_matches[match_lang]->MIME_name ?
					d_matches[match_lang]->MIME_name :
					"<no MIME name>"));
		    }
		}
	    }

	    for (j = 0; j < MATCH_COUNT; j++)
		if (d_matches[j]) {
		    system_charset = d_matches[j];
		  
		    DPRINT(Debug,3,(&Debug,
				    "map_system_charset: using match %d: \"%s\" (codeset \"%s\") as %s\n", 
				    j,lc_ctype,codeset,
				    system_charset->MIME_name ?
				    system_charset->MIME_name :
				    "<no MIME name>"));
		    return;
		}
	} else {
	    int val;
	    char buffer[STRING];
	
	    if (d_matches[match_codeset]) {
		system_charset = d_matches[match_codeset];
		DPRINT(Debug,3,(&Debug,
				"map_display_charset: using match codeset: \"%s\" (codeset \"%s\") as %s\n", 
				lc_ctype,codeset,
				system_charset->MIME_name ?
				system_charset->MIME_name :
				"<no MIME name>"));
		return;
	    }
  
	    /*  1) ------ guess based on codeset ------------------------ */

	    if (codeset[0] &&
		     0 == strincmp(codeset,"ISO-8859-",9) &&
		     0 < (val = atoi(codeset+9))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else if (codeset[0] &&
		       0 == strincmp(codeset,"ISO_8859-",9) &&
		       0 < (val = atoi(codeset+9))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else if (codeset[0] &&
		       0 == strincmp(codeset,"ISO8859-",8) &&
		       0 < (val = atoi(codeset+8))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else if (codeset[0] &&
		       0 == strincmp(codeset,"ISO8859",7) &&
		       isascii(codeset[7]) && isdigit(codeset[7]) && 
		       '\0' == codeset[8]) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%c"),
			     codeset[7]);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);

	    /*  2) ------ guess based on lc_ctype ------------------------ */

	    } else if (0 == strcmp(lc_ctype,"C"))
		system_charset = MIME_name_to_charset("US-ASCII",
						      CHARSET_create); 
	    else if (0 == istrcmp(lc_ctype,"ASCII") ||
		     0 == istrcmp(lc_ctype,"US-ASCII"))
		system_charset = MIME_name_to_charset("US_ASCII",
						      CHARSET_create);
	    else if (0 == strincmp(lc_ctype,"ISO-8859-",9) &&
		     0 < (val = atoi(lc_ctype+9))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,CHARSET_create);
	    } else if (0 == strincmp(lc_ctype,"ISO_8859-",9) &&
		       0 < (val = atoi(lc_ctype+9))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,CHARSET_create);
	    } else if (0 == strincmp(lc_ctype,"ISO8859-",8) &&
		       0 < (val = atoi(lc_ctype+8))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,CHARSET_create);

	    /*  3) ------ guess based on charset part of locale ------------------------ */

	    } else if (charset && (0 == istrcmp(charset,"ASCII") ||
				   0 == istrcmp(charset,"US-ASCII")))
		system_charset = MIME_name_to_charset("US-ASCII",
						      CHARSET_create);
	    else if (charset &&
		     0 == strincmp(charset,"ISO-8859-",9) &&
		     0 < (val = atoi(charset+9))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else if (charset &&
		       0 == strincmp(charset,"ISO_8859-",9) &&
		       0 < (val = atoi(charset+9))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else if (charset &&
		       0 == strincmp(charset,"ISO8859-",8) &&
		       0 < (val = atoi(charset+8))) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%d"),
			     val);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else if (charset &&
		       0 == strincmp(charset,"ISO8859",7) &&
		       isascii(charset[7]) && isdigit(charset[7]) && 
		       '\0' == charset[8]) {
		elm_sfprintf(buffer, sizeof buffer,
			     FRM("ISO-8859-%c"),
			     charset[7]);
		system_charset = MIME_name_to_charset(buffer,
						      CHARSET_create);
	    } else {
	    
		if (global_only) {

		    DPRINT(Debug,3,(&Debug,
				    "map_display_charset: global_only=%d: Failed silently\n",
				    global_only));

		} else {
  
		    lib_error(FRM("Elm: Unable to map %.20s locale (LC_CTYPE) to MIME charset."),
			      lc_ctype);
		    
		    if (codeset[0]) {
			system_charset = MIME_name_to_charset(codeset,
							      CHARSET_create);
			lib_error(FRM("     Codeset name %s was unknown. Treating %s as MIME name."),
				  codeset,codeset);
			lib_error(FRM("     Check %.100s or %.100s"),
				  system_mime_charsets, user_mime_charsets);
			
		    } else {
			system_charset = MIME_name_to_charset(default_system_charset,
							      CHARSET_create);
			lib_error(FRM("     Using default (%.20s). Check %.100s or\n      %.100s"),
				  default_system_charset,system_mime_charsets,
				  user_mime_charsets);
		    }
		    locale_error = 1;
		}
	    }

	    DPRINT(Debug,3,(&Debug,
			    "map_display_charset: default \"%s\" as \"%s\"\n", 
			    lc_ctype, system_charset->MIME_name));
	}
    }
}

static int user_rc_set = 0;
/* shell value from passwd */
static char shell0[STRING] = "";

void user_init() 
{
    char *cp;
    int r;
    
    /** initialize the whole ball of wax.
     **/
    struct passwd *pass;
    
    if (!locale_init_called)
	locale_init ();
    
    /* save original user and group ids */
    userid  = getuid();
    groupid = getgid();	
    
    /* GNU library seems give _SC_SAVED_IDS as enumerated
       value instead of #define so we can not test if
       it is defined (as macro)
    */
#if defined(_SC_SAVED_IDS) || defined(_POSIX_VERSION)
    r = sysconf(_SC_SAVED_IDS);
    if (r < 0) {
	int err UNUSED_VAROK = errno;
	DPRINT(Debug,1,(&Debug,
			"sysconf(_SC_SAVED_IDS) failed: (errno=%d) %s\n",
			err,strerror(err)));	
	    
	have_saved_ids = 0;
    } else {
	have_saved_ids = r > 0;
	DPRINT(Debug,1,(&Debug,
			"*** have_saved_ids=%d\n",have_saved_ids));
    }
#else
    have_saved_ids = 0;
#endif
    mailgroupid = getegid();
    
    /* Get username (logname), home (login directory), and full_username
     * (part of GCOS) field from the password entry for this user id.
     * Full_username will get overridden by fullname in elmrc, if defined.
     *
     * For those sites that have various user names with the same user
     * ID, use the passwd entry corresponding to the user name as long 
     * as it matches the user ID.  Otherwise fall back on the entry 
     * associated with the user ID alone.
     */

    if((cp = getenv("LOGNAME")) == NULL)
	cp = getenv("USER");

    errno = 0;             /* Errno is not necessary set on error */

    if(cp != NULL && (pass = getpwnam(cp)) != NULL &&
       pass->pw_uid == userid) {

	DPRINT(Debug,1,(&Debug,"Found username %s ($LOGNAME or $USER)\n",
			cp));

    } else if((pass = getpwuid(userid)) == NULL) {
	int err = errno;

	if (err) 
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmPasswordEntryError,
			      "Failed to get your password entry: %s"),
		      strerror(err));
	else	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmYouHaveNoPasswordEntry,
			      "You have no password entry!"));

	exit(1);
    }

    strfcpy(username, pass->pw_name, sizeof username);

    if (pass->pw_dir)
	strfcpy(home, pass->pw_dir, sizeof home);
    else if (NULL != (cp = getenv("HOME")))
	strfcpy(home,cp,sizeof home);
    else
	strfcpy(home,"/",sizeof home);
    
    if (pass->pw_shell)
	strfcpy(shell0, pass->pw_shell, sizeof shell0);

    /* Null shell in /etc/passwd means /bin/sh */
    if ('\0' == shell0[0]) 
	strfcpy(shell0, "/bin/sh", sizeof shell0);

    /* utils/from.c calls this again for resetting */
    if (full_username_s) 
	free_string(&full_username_s);

    if((cp = get_fullname1(pass, username)) != NULL)
	full_username_s = new_string2(system_charset,s2us(cp));
    else
	full_username_s = new_string2(system_charset,s2us(username));
    /* fall back on logname */


    if (!user_rc_set)
	elm_sfprintf(user_rc_file, sizeof user_rc_file,
		     FRM("%s/%s"), home, ELMRCFILE);

    elm_sfprintf(old_user_text_file, sizeof old_user_text_file,
		 FRM("%s/%s"), home, OLD_USER_ALIAS_TEXT);
    elm_sfprintf(old_user_data_file, sizeof old_user_data_file,
		 FRM("%s/%s"), home, OLD_USER_ALIAS_DATA);
    elm_sfprintf(user_aliases_file, sizeof user_aliases_file,
		 FRM("%s/%s"), home, USER_ALIASES_FILE);

    elm_sfprintf(user_mime_types, sizeof user_mime_types,
		 FRM("%s/%s"), home, USER_MIME_TYPES);
    elm_sfprintf(user_mailheaders, sizeof user_mailheaders,
		 FRM("%s/%s"), home, MAILHEADERS);
    elm_sfprintf(user_mime_charsets, sizeof user_mime_charsets,
		 FRM("%s/%s"), home, USER_MIME_CHARSETS);
    elm_sfprintf(user_iso2022_sets, sizeof user_iso2022_sets,
		 FRM("%s/%s"), home, USER_ISO2022_SETS);
    elm_sfprintf(user_terminal_info, sizeof user_terminal_info,
		 FRM("%s/%s"), home, USER_TERMINAL_INFO);
    elm_sfprintf(user_mail_services, sizeof user_mail_services,
		 FRM("%s/%s"), home, USER_MAIL_SERVICES);
    elm_sfprintf(user_mlist_file, sizeof user_mlist_file,
		 FRM("%s/%s"), home, USER_MLIST_FILE);


#if DEBUG    
    /* Open debug files and print spooled debug output */
    debug_user_init();
   
    DPRINT(Debug,1,(&Debug,"  username = %s\n",username));
    DPRINT(Debug,1,(&Debug,"  home     = %s\n",home));
    if (full_username_s) {
	DPRINT(Debug,1,(&Debug,"  fullname = %S\n",full_username_s));
    }

#endif
}

void set_user_rc_location(p)
     char *p;
{
    strfcpy(user_rc_file, p, sizeof user_rc_file);
    user_rc_set = 1;
}


#ifdef I_NETDB
/* Seems that h_errno is macro on AIX */
#ifndef h_errno
extern int h_errno;
#endif
#endif

#ifdef I_NETINET_IN
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

static int lookup_domain_name P_((char * name,const char ***haliases));
static int lookup_domain_name(name,haliases)
     char * name;
     const char ***haliases;
{

#define MAX_IP  10
#define IPNUMLEN  (4*4)

    static char name_buffer[MAX_IP][IPNUMLEN+3];

    if (haliases && *haliases)
	(*haliases)[0] = NULL;
   
#ifdef I_NETDB
    
    if (name) {
	struct hostent *he = gethostbyname(name);
	if (!he) {
	    DPRINT(Debug,3,(&Debug,
			    "lookup_domain_name: %s: h_errno=%d\n",
			    name,h_errno));
	    switch(h_errno) {
	    case TRY_AGAIN:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeTemporary,
				  "Address for %s is not yet found..."),
			  name);
	    }
	} else {
	    int ok = 0;
	    int idx;
	    char *p;

	    int count = 0;


	    if (NULL != (p = 
			 strchr(he->h_name ,'.'))) {
		if (!hostdomain[0])
		    strfcpy(hostdomain,p,sizeof hostdomain);
		if (!hostfullname[0]) 
		    strfcpy(hostfullname, he->h_name, sizeof hostfullname);
		ok = 1;

	    } else if (!hostname[0]) {
		strfcpy(hostname, he->h_name, sizeof hostname);
		DPRINT(Debug,3,(&Debug,
				"lookup_domain_name: \"%s\" -> hostname=\"%s\"\n",
				name,hostdomain,hostname));
	    }

	    if (haliases) {
		*haliases = safe_realloc(*haliases,
					 2 * sizeof ((*haliases)[0]));

		(*haliases)[0] = he->h_name;
		count = 1;
	    }

	    for (idx = 0; he->h_aliases[idx]; idx++) {
		if (!ok &&
		    NULL != (p = 
			     strchr(he->h_aliases[idx],'.'))) {
		    if (!hostdomain[0])
			strfcpy(hostdomain,p,sizeof hostdomain);
		    if (!hostfullname[0]) 
			strfcpy(hostfullname, he->h_aliases[idx], 
				sizeof hostfullname);
		    ok = 1;
		} 

		if (haliases) {
		    *haliases = safe_realloc(*haliases,
					     ( count + 2) * sizeof ((*haliases)[0]));
		    
		    (*haliases)[count] = he->h_aliases[idx];
		    count++;
		    
		}
	    }
	    if (ok) {
		DPRINT(Debug,3,(&Debug,
				"lookup_domain_name: \"%s\" -> hostdomain=\"%s\", hostfullname=\"%s\"\n",
				name,hostdomain,hostfullname));
	    }

	    if (haliases) {
		for (idx = 0; he->h_addr_list[idx] && idx < MAX_IP; idx++) {
	
		    switch(he->h_addrtype) {
		       
#ifdef I_NETINET_IN
		    case AF_INET: {
			struct sockaddr_in      sin;    /* INET family */
						
			if (sizeof(sin.sin_addr)  != he->h_length) {
			    lib_error(FRM("%s: Bad IP-addr length %d (should be %d)"),
				      he->h_name,he->h_length,
				      sizeof(sin.sin_addr));
			    break;
			}
			memcpy(&(sin.sin_addr), he->h_addr_list[idx], he->h_length);

			elm_sfprintf(name_buffer[idx], sizeof name_buffer[idx],
				     FRM("[%s]"),inet_ntoa( sin.sin_addr));

			*haliases = safe_realloc(*haliases,
						 ( count + 2) * sizeof ((*haliases)[0]));
			
			(*haliases)[count] = name_buffer[idx];
			count++;
			
		    }   
			break;
#endif
		    }
		}
	    }

	    if (haliases && *haliases ) {
		int i;

		(*haliases)[count] = NULL;
		
		for (i = 0; i < count; i++) {
		    DPRINT(Debug,7,(&Debug, 
				    "lookup_domain_name:  \"%s\" -> host alias[%d]=\"%s\"",
				    name,i,(*haliases)[i]));
		}
	    }

	    DPRINT(Debug,7,(&Debug,
			    "lookup_domain_name=%d\n",
			    ok));

	    return ok;
	}
    }
#endif    

    return 0;
}

static void generate_hostfullname P_((void));
static void generate_hostfullname() 
{
    int hostlen, domlen;

    /*
     * now the tough part:
     *	we need to make three variables out of this stuff:
     *	hostname = just the hostname, as in bangpaths,
     *		this is whatever the user gave us so far,
     *		we wont change this one
     *	hostdomain = this is the domain considered local to this
     *		machine, and should be what we got above.
     *	hostfullname = this is the full FQDN of this machine,
     *		and is a strange combination of the first two.
     *	if tail(hostname) == hostdomain
     *		then hostfullname = hostname
     *			ie: node.ld.domain.type, ld.domain.type -> node.ld.domain.type
     *	else if hostname == hostdomain + 1
     *		then hostfullname = hostname
     *			ie: domain.type, .domain.type -> domain.type
     *	
     *	else hostfullname = hostname + hostdomain
     *			ie: host, .domain.type -> host.domain.type
     * lost yet?
     */
    
    DPRINT(Debug,3,(&Debug,
		    "hostname=\"%s\",hostdomain=\"%s\" ",
		    hostname,hostfullname));
    
    hostlen = strlen(hostname);
    domlen = strlen(hostdomain);
    if (hostlen >= domlen) {
	if (istrcmp(&hostname[hostlen - domlen], hostdomain) == 0)
	    strfcpy(hostfullname, hostname, sizeof hostfullname);
	else {
	    strfcpy(hostfullname, hostname, sizeof hostfullname);
	    strfcat(hostfullname, hostdomain, sizeof hostfullname);
	}
    } else {
	if (istrcmp(hostname, hostdomain + 1) == 0)
	    strfcpy(hostfullname, hostname, sizeof hostfullname);
	else {
	    strfcpy(hostfullname, hostname, sizeof hostfullname);
	    strfcat(hostfullname, hostdomain, sizeof hostfullname);
	}
    }
    
    DPRINT(Debug,3,(&Debug,
		    "-> hostfullname=\"%s\"\n",
		    hostfullname));
}

static int init_defaults_called = 0;

static init_default_hook **hooklist = NULL;
static int                hooklist_len = 0;

void add_init_default_hook(hook)
     init_default_hook *hook;
{

    if (init_defaults_called) {
	hook(&errors);
    }

    hooklist = safe_realloc(hooklist,(hooklist_len+1) * sizeof(hooklist[0]));

    hooklist[hooklist_len] = hook;
    hooklist_len++;
}

static unsigned long name_resolution_val = 0L;

int have_name_resolution(tag)
     enum name_resolution tag;
{
    if (tag < 0 || tag >= nr_COUNT)
	panic("RC PANIC",__FILE__,__LINE__,"have_name_resolution",
	      "Bad tag", 0);
    
    return 0 != ( name_resolution_val & ( 1 << tag ));
}

#if ANSI_C
struct name_resolution_rec;
#endif
typedef void name_resolution_init_f P_((struct name_resolution_rec *rec,
					enum name_resolution tag));

S_(name_resolution_init_f nr_null_init)
static void nr_null_init P_((struct name_resolution_rec *rec,
			  enum name_resolution tag));
static void nr_null_init(rec,tag)
     struct name_resolution_rec *rec;
     enum name_resolution tag;
{
    /* nothing */
}

S_(name_resolution_init_f nr_not_avail)
static void nr_not_avail P_((struct name_resolution_rec *rec,
			     enum name_resolution tag));
static void nr_not_avail(rec,tag)
     struct name_resolution_rec *rec;
     enum name_resolution tag;
{
    /* nothing */
}

S_(name_resolution_init_f nr_GETHOSTNAME)
static void nr_GETHOSTNAME P_((struct name_resolution_rec *rec,
			       enum name_resolution tag));
S_(name_resolution_init_f nr_UNAME)
static void nr_UNAME P_((struct name_resolution_rec *rec,
			 enum name_resolution tag));
S_(name_resolution_init_f nr_GETDOMAINNAME)
static void nr_GETDOMAINNAME P_((struct name_resolution_rec *rec,
				 enum name_resolution tag));
#ifdef MAILNAME_FILE
S_(name_resolution_init_f nr_MAILNAME)
static void nr_MAILNAME P_((struct name_resolution_rec *rec,
			    enum name_resolution tag));
#endif
static struct name_resolution_rec {
    const char * tag;
    
    char * result;
    int errno_val;
    name_resolution_init_f *init;
} name_resolutions[nr_COUNT] = {  
    /* NOTE: Must be on same order than on enum name_resolution */

    /* nr_gethostname */
    { "gethostname", NULL, 0,
#ifdef GETHOSTNAME
      nr_GETHOSTNAME
#else
      nr_not_avail
#endif
    },

    /* nr_uname */
    { "uname", NULL, 0,
#ifdef DOUNAME
      nr_UNAME
#else
      nr_not_avail
#endif
    },

    /* nr_lookup */
    { "lookup", NULL, 0,
#ifdef I_NETDB     
      nr_null_init
#else
      nr_not_avail
#endif
    },

    /* nr_getdomainname */
    { "getdomainname", NULL, 0,
#ifdef GETDOMAINNAME
      nr_GETDOMAINNAME
#else
      nr_not_avail
#endif
    },

    /* nr_mailname */
    { "mailname-file", NULL, 0,
#ifdef MAILNAME_FILE
      nr_MAILNAME
#else
      nr_not_avail
#endif
    },
};

const char * name_resolution_tag(tag)
     enum name_resolution tag;
{
    if (tag < 0 || tag >= nr_COUNT)
	panic("RC PANIC",__FILE__,__LINE__,"name_resolution_tag",
	      "Bad tag", 0);
    
    return name_resolutions[tag].tag;
}

const char * name_resolution_gives(tag)
     enum name_resolution tag;
{
    if (tag < 0 || tag >= nr_COUNT)
	panic("RC PANIC",__FILE__,__LINE__,"name_resolution_tag",
	      "Bad tag", 0);
    
    return name_resolutions[tag].result;
}

int name_resolution_errno(tag)
     enum name_resolution tag;
{
    if (tag < 0 || tag >= nr_COUNT)
	panic("RC PANIC",__FILE__,__LINE__,"name_resolution_errno",
	      "Bad tag", 0);

    return name_resolutions[tag].errno_val;
}

#ifdef GETHOSTNAME
static void nr_GETHOSTNAME(rec,tag) 
     struct name_resolution_rec *rec;
     enum name_resolution tag;
{
    int len = SLEN;
    char * buffer;

#if HOST_NAME_MAX
    if (len < HOST_NAME_MAX)
	len = HOST_NAME_MAX;
#endif    

    buffer = safe_malloc(len+1);
    bzero(buffer,len+1);
    
    /* Truncation of result may occur without error */

    if (gethostname(buffer,len) < 0) {
	int err = errno;
	DPRINT(Debug,8,(&Debug,
			"gethostname failed: (errno=%d) %s\n",
			err,strerror(err)));	

	rec->errno_val = err;

	/* glibc uses ENAMETOOLONG */
#ifdef 	 ENAMETOOLONG
	if (ENAMETOOLONG == err)
	    goto truncated;
#endif

    } else {
    truncated:
	buffer[len] = '\0';

	DPRINT(Debug,8,(&Debug,
			"gethostname gives: %s\n",buffer));

	if (buffer[0])
	    rec->result = strmcpy(rec->result,buffer);

    }

    free(buffer);
}
#endif

#ifdef DOUNAME
static void nr_UNAME(rec,tag) 
     struct name_resolution_rec *rec;
     enum name_resolution tag;
{
    struct utsname u_name;

    if (uname(&u_name) < 0) {
	int err = errno;
	DPRINT(Debug,1,(&Debug,
			"uname failed: (errno=%d) %s\n",
			err,strerror(err)));	
	rec->errno_val = err;

    } else {
	DPRINT(Debug,8,(&Debug,
			"uname gives: %s\n",u_name.nodename));

	if (u_name.nodename[0])
	    rec->result = strmcpy(rec->result,u_name.nodename);
    }
}
#endif

#ifdef GETDOMAINNAME
static void nr_GETDOMAINNAME(rec,tag)
     struct name_resolution_rec *rec;
     enum name_resolution tag;
{
    int len = SLEN;
    char * buffer;

#if HOST_NAME_MAX
    if (len < HOST_NAME_MAX)   /* Not correct */
	len = HOST_NAME_MAX;
#endif    
    
    buffer = safe_malloc(len+1);
    bzero(buffer,len+1);
    
    /* Can truncation of result occur without error ? */

    if (getdomainname(buffer,len) < 0) {
	int err = errno;
	DPRINT(Debug,8,(&Debug,
			"getdomainname failed: (errno=%d) %s\n",
			err,strerror(err)));	

	rec->errno_val = err;

	/* ENAMETOOLONG not used ? */
#ifdef 	 ENAMETOOLONG
	if (ENAMETOOLONG == err)
	    goto truncated;
#endif
		
    } else {
    truncated:
	buffer[len] = '\0';

	DPRINT(Debug,8,(&Debug,
			"getdomainname gives: %s\n",buffer));

	if (buffer[0])
	    rec->result = strmcpy(rec->result,buffer);

    }

    free(buffer);
}
#endif

#ifdef MAILNAME_FILE
static void nr_MAILNAME(rec,tag)
     struct name_resolution_rec *rec;
     enum name_resolution tag;
{
    int len = SLEN;

#if HOST_NAME_MAX
    if (len < HOST_NAME_MAX)   /* Not correct */
	len = HOST_NAME_MAX;
#endif    
    FILE    *fp = fopen(MAILNAME_FILE,"r");

    if (fp) {
	int r = malloc_gets(& (rec->result), len, fp);

	if (-1 == r) {
	    DPRINT(Debug,8,(&Debug,
			    "%s truncated (max=%d): %s\n",
			    MAILNAME_FILE,len,
			    rec->result ? rec->result : "(none)"));
	} else if (-1 == r) {
	    int err = errno;
	    DPRINT(Debug,8,(&Debug,
			    "%s read failed: (errno=%d) %s\n",
			    MAILNAME_FILE,err,strerror(err)));
	    rec->errno_val = err;

	} else if (0 == r) {
	    DPRINT(Debug,8,(&Debug,
			    "%s is empty\n",MAILNAME_FILE));

	}

	if (rec->result) {
	    DPRINT(Debug,8,(&Debug,
			    "%s gives: %s\n",MAILNAME_FILE,rec->result));
	}

	fclose(fp);
    } else {
	int err = errno;
	DPRINT(Debug,8,(&Debug,
			"%s open failed: (errno=%d) %s\n",
			MAILNAME_FILE,err,strerror(err)));
	rec->errno_val = err;

    }
}
#endif

/* static pointer to buffer accross invocations */
static char * nameres_return_buffer = NULL;

int nameresfunc(value,enter,lineno,filename)
     char **value; 
     int enter;
     int lineno; 
     const char *filename;
{
    static unsigned long incremental_add = 0L;
    static unsigned long incremental_sub = 0L;

    int ok = 1;    

    if (enter) {
	char *ptr, *next = NULL;
	int absolute = 0;

	incremental_add = 0L;
	incremental_sub = 0L;
	
	for (ptr = *value; ptr && *ptr; ptr = next) {
	    int len;
	    int mode = 0;
	    enum name_resolution tag;

	    while (whitespace(*ptr))
		ptr++;

	    next = strpbrk(ptr,"\t #");

	    if (!next)
		len = strlen(ptr);
	    else {
		len = next - ptr;
		
		if ('#' == *next)
		    next = NULL;
	    }

	    if (len < 1)
		continue;
	    
	    if ('+' == *ptr || '-' == *ptr) {
		mode = *ptr;
		ptr++;
		len--;
	    }

	    if (mode && absolute) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNameResolutionMix,
				  "Mixing of incremental mode and absolute mode on \"name-resolution\" in line %d in \"%s\" file is not supported: %c%.*s"),
			  lineno,filename,mode,len,ptr);
		ok = 0;

	    } else if (!mode && !absolute) {
		
		if (incremental_add || incremental_sub) {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmNameResolutionMix1,
				      "Mixing of incremental mode and absolute mode on \"name-resolution\" in line %d in \"%s\" file is not supported: %.*s"),
			      lineno,filename,len,ptr);
		    ok = 0;
		}

		absolute = 1;
		name_resolution_val = 0L;

		/* This is only valid as first value */
		if (0 == strcmp(ptr,"none")) 		    
		    continue;                   		
	    }
		
	    for (tag = 0; tag < nr_COUNT; tag++) {
		const char * T = name_resolution_tag(tag);

		if (0 == strncmp(ptr,T,len) && '\0' == T[len])
		    break;
	    }
	    
	    if (tag < nr_COUNT) {
		unsigned long val = 1 << tag;
		
		if (mode != '-' && 
		    nr_not_avail == name_resolutions[tag].init) {

		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmNameResolutionUnsupportedkw,
				      "Unsupported keyword on \"name-resolution\" in line %d in \"%s\" file: %.*s"),
			      lineno,filename,len,ptr);
		    ok = 0;	

		} else { 
		    switch (mode) {			
		    case '+':
			incremental_add      |= val;
			name_resolution_val  |= val;
			break;
		    case '-':
			incremental_sub      |= val;
			name_resolution_val  &= ~val;
			break;
		    case 0:
			name_resolution_val  |= val;
			break;
		    }
		}
	    } else {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNameResolutionBadkw,
				  "Bad keyword on \"name-resolution\" in line %d in \"%s\" file: %.*s"),
			  lineno,filename,len,ptr);
		ok = 0;	
	    }
	}
       
    } else {
	enum name_resolution tag;

	if (nameres_return_buffer)
	    free(nameres_return_buffer);
	nameres_return_buffer = NULL;

	if (incremental_add || incremental_sub) {

	    for (tag = 0; tag < nr_COUNT; tag++) {
		const char * T = name_resolution_tag(tag);
		unsigned long val = 1 << tag;

		if (val == (incremental_add & val)) {

		    char * buf = 
			elm_message(FRM("%s%s+%s"),
				    nameres_return_buffer ? nameres_return_buffer : "",
				    nameres_return_buffer ? " " : "",
				    T);
		    
		    if (nameres_return_buffer)
			free(nameres_return_buffer);
		    nameres_return_buffer = buf;		    
		} 

		if (val == (incremental_sub & val)) {
		    
		    char * buf = 
			elm_message(FRM("%s%s-%s"),
				    nameres_return_buffer ? nameres_return_buffer : "",
				    nameres_return_buffer ? " " : "",
				    T);
		    
		    if (nameres_return_buffer)
			free(nameres_return_buffer);
		    nameres_return_buffer = buf;		    
		} 
	    }

	    /* If on incremental mode, show real value after comment */

	    nameres_return_buffer = 
		strmcat(nameres_return_buffer," # ");
	}

	if (!name_resolution_val && !nameres_return_buffer) {
	    *value = "none";
	} else {

	    if (!name_resolution_val) {
		nameres_return_buffer = 
		    strmcat(nameres_return_buffer," none");
	    }
		
	    for (tag = 0; tag < nr_COUNT; tag++) {
	
		if (have_name_resolution(tag)) {
		    const char * T = name_resolution_tag(tag);
		    char * buf = 
			elm_message(FRM("%s%s%s"),
				    nameres_return_buffer ? nameres_return_buffer : "",
				    nameres_return_buffer ? " " : "",
				    T);
		    
		    if (nameres_return_buffer)
			free(nameres_return_buffer);
		    nameres_return_buffer = buf;		    		    
		}
	    }

	    *value = nameres_return_buffer;	    
	}
    }
   
    return ok;
}

/* Returns name only if it is ok, may update error counter and print error message */

const char * name_resolution_okname(tag,global_mode)
     enum name_resolution tag;
     int global_mode;
{
    const char * R = name_resolution_gives(tag);
    const char * T = name_resolution_tag(tag);
    const unsigned char *x;
    int l;

    if (!R)
	return NULL;

    if (! R[0])
	return NULL;

    /* For some reason getdomainname() returns "(none)" 
       Silently ignore it. 
      
       ( ) are comment characters on email addresses
    */
    if ('(' == R[0] && (2 < (l = strlen(R))) && ')' == R[l-1]) {
	DPRINT(Debug,4,(&Debug,
			"name_resolution_okname(%d,%d): %s: ignoting value %s\n",
			tag,global_mode,T,R));

	return NULL;
    }

    for (x = cs2us(R); *x; x++) {
	if (!isascii(*x)) { 
	    
	    if (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 
		0 == access(system_rc_file,WRITE_ACCESS)) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNameAsciiOnly,
				  "Only ASCII characters (not 0x%02x) are allowed on value \"%s\" given by %s"),
			  *x,R,T);
		errors++;	
	    }

	    return NULL;
	}
	
	if (!isgraph(*x)) {
	    if (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 
		0 == access(system_rc_file,WRITE_ACCESS)) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNameGraphOnly,
				  "Only visible characters (not 0x%02x) are allowed on value \"%s\" given by %s"),
			  *x,R,T);
		errors++;	
	    }

	    return NULL;
	}
	
	/* '.' is only RFC 822 special which is allowed */
	if ('.' != *x &&
	    rfc822_special(*x)) {

	    if (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 
		0 == access(system_rc_file,WRITE_ACCESS)) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNameNormalOnly,
				  "Only normal characters (not special %c) are allowed on value \"%s\" given by %s"),
			  *x,R,T);
		errors++;	
	    }
	    
	    return NULL;
	}
	    
    }

    return R;
}

static void add_no_duplicate P_((struct dt_path_info *ptr,
				 const char *fieldname,
				 const char *val));
static void add_no_duplicate(ptr,fieldname,val)
     struct dt_path_info *ptr; 
     const char *fieldname; 
     const char *val;
{
    const char ** list = 
	give_dt_path_as_elems(ptr,fieldname);
    
    if (list) {
	int i;
	for (i = 0; list[i]; i++) 
	    if (0 == strcmp(list[i],val))
		return;
    }
    
    /* Fake recursive call */
	
    expand_dt_path(ptr,fieldname,val,2);
}
static void setup_names P_((int rc_open_errno, int global_mode));
static void setup_names(rc_open_errno,global_mode)
     int rc_open_errno;
     int global_mode;
{
    const char **haliases = NULL; 
    
    const char ** mailname_list = 
	give_dt_path_as_elems(&mailname,"mailname");

    /*
     * Get the host name as per configured behavior.
     */

    /* hostdomain, hostfullname, hostname, name-resolution, mailname
       are set with FL_ASCIIFIRST so they are set when this is called
    */

    if (!hostname[0]) {

	enum name_resolution vector[] = { nr_gethostname, nr_uname, nr_BAD };
	int i;

	for (i = 0; vector[i] != nr_BAD; i++) {
	    enum name_resolution tag = vector[i];

	    if (have_name_resolution(tag)) {
		const char * R = name_resolution_okname(tag,global_mode);
		const char * T = name_resolution_tag(tag);
		int E = name_resolution_errno(tag);
		
		if (E && (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 
			  0 == access(system_rc_file,WRITE_ACCESS))) {
		    
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmHostnameErrno,
				      "Hostname query by %s failed: %s"),
			      T,strerror(E));
		    
		    errors++;		
		}
		
		if (R) {
		    int L;
		    strfcpy(hostname,R, sizeof hostname);
		    
		    if ((L = strlen(R)) > sizeof  hostname -1 &&
			(0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 
			 0 == access(system_rc_file,WRITE_ACCESS))) {
			
			lib_error(CATGETS(elm_msg_cat, ElmSet, 
					  ElmHostnameTooLong,
					  "Hostname given by %s is too long (%d, max %d): %s"),
				  T,L,sizeof  hostname -1, R);
			
			errors++;		       
		    }
		    
		    break;
		}
	    }
	}
    }
    
    if (hostfullname[0] && !hostdomain[0]) {
	char * p = strchr(hostfullname,'.');

	if (p && NULL != strchr(p+1,'.')) {
	    
	    strfcpy(hostdomain,p,sizeof hostdomain);

	    DPRINT(Debug,3,(&Debug,
			    "hostfullname=\"%s\" -> hostdomain=\"%s\"\n",
			    hostfullname,hostdomain));

	    if (!hostname[0]) {
		int L = p - hostfullname;

		strfcpy(hostname,hostfullname,sizeof hostname);
		if (L < sizeof hostname)
		    hostname[L] = '\0';

		DPRINT(Debug,3,(&Debug,
				" ... -> hostname=\"%s\"\n",
				hostname));

	    }
	}	
    }

    if ((hostname[0] || hostfullname[0]) &&
	have_name_resolution(nr_lookup)) {
	int fqdn_found UNUSED_VAROK = 0;

	if (!hostdomain[0] || !hostfullname[0] || ! mailname_list) {

	    if (hostfullname[0])
		fqdn_found = lookup_domain_name(hostfullname,&haliases);
	    else if (hostname[0])
		fqdn_found = lookup_domain_name(hostname,&haliases);
	}
    }

    /*
     * now get the domain name, used to build the full name
     */
    if (!hostdomain[0]) {
	if (gethostdomain(hostdomain, sizeof(hostdomain), global_mode) <0) {
	    DPRINT(Debug,1,(&Debug, 
			    "Default domain name is not available\n"));

	    hostdomain[0] = '\0';
	}

    } else {
	if ('.' != hostdomain[0]) {
	    char buffer[SLEN];
	    strfcpy(buffer,hostdomain,sizeof buffer);
	    strfcpy(hostdomain,".",sizeof hostdomain);
	    strfcat(hostdomain,buffer,sizeof hostdomain);
	    DPRINT(Debug,3,(&Debug,
			    "hostdomain=\"%s\" -> hostdomain=\"%s\"\n",
			    buffer,hostdomain));
	}
    }

    if (hostname[0] && hostdomain[0] && !hostfullname[0]) {
	generate_hostfullname();

    }

    /* If global elm.rc gives
            mailname = none
       then
           mailname_list != NULL 
       but
           mailname_list[0] == NULL
    */
    
    if (mailname_list && mailname_list[0] && NULL == strchr(mailname_list[0],'.')) {
	int i;
	
	if (hostfullname[0]) {
	    char *x;
	    int dot = 0;

	    for (x = hostfullname; *x; x++) {
		if (!isascii(*x)) 
		    break;
		if ('.' == *x)
		    dot = 1;
	    }

	    if (! *x && dot) {

		if (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 0 == access(system_rc_file,WRITE_ACCESS)) {

		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailnameQualified,
				      "First hostname on \"mailname\" on RC file %s should be qualified: %s"),
			      system_rc_file,mailname_list[0]);
		    
		    errors++;		       
		}

		mailname_list = NULL;  /* Returned pointer is invalid after reallocation */
		
		mailname.list = safe_realloc(mailname.list,
					     (mailname.nlen+2)* sizeof (mailname.list[0]));
		
		for (i = mailname.nlen; i > 0; i--)
		    mailname.list[i] = mailname.list[i-1];
		mailname.list[0] = safe_strdup(hostfullname);
		mailname.nlen++;
		mailname.list[mailname.nlen] = NULL;
		
		if (mailname.unexpanded)
		    free(mailname.unexpanded);
		mailname.unexpanded = NULL;
	    } else
		goto print_mailname_error;

	} else {
	print_mailname_error:

	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailnameQualified,
			      "First hostname on \"mailname\" on RC file %s should be qualified: %s"),
		      system_rc_file,mailname_list[0]);
	    
	    errors++;		       
	}


    } else if (! mailname_list && 
	       ( hostfullname[0] || 
		 (have_name_resolution(nr_mailname) && 
		  name_resolution_okname(nr_mailname,
					 global_mode))
		 )
	       ) { 

	/* nr_BAD             is -1 
	   nr_DEFAULT_names   is -2
	   nr_DEFAULT_aliases is -3
	*/
		
	int vector[] = { nr_mailname, nr_DEFAULT_names, 
			 nr_gethostname, nr_uname, 
			 nr_DEFAULT_aliases, nr_BAD };
	int i;
		   
	for (i = 0; vector[i] != nr_BAD; i++) {
	    enum name_resolution tag = vector[i];
	
	    switch (tag) {

	    case nr_DEFAULT_names:

		if (hostfullname[0])
		    add_no_duplicate(&mailname,"mailname",hostfullname);
	
		if (hostname[0])
		    add_no_duplicate(&mailname,"mailname",hostname);
	
		break;
		
	    case nr_DEFAULT_aliases:
		if (haliases) {
		    int j;

		    for (j = 0; haliases[j]; j++) {
			add_no_duplicate(&mailname,"mailname",haliases[j]);	
		    }
		}
		break;

	    default:
		if (have_name_resolution(tag)) {
		    const char * R = name_resolution_okname(tag,global_mode);
		    const char * T = name_resolution_tag(tag);
		    int E = name_resolution_errno(tag);
		    
		    if (E && (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 
			      0 == access(system_rc_file,WRITE_ACCESS))) {
			
			lib_error(CATGETS(elm_msg_cat, ElmSet, 
					  ElmMailnameErrno,
					  "Mailname query by %s failed: %s"),
				  T,strerror(E));
			
			errors++;		
		    }
		    
		    if (R) {
			add_no_duplicate(&mailname,"mailname",R);	
		    }
		}
	    }
	}
    }

    
    if (!hostname[0] || !hostdomain[0] || !hostfullname[0]) {
	
	if (rc_open_errno) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedSystemRC,
			      "Failed to open RC file %s: %s"),
		      system_rc_file,strerror(rc_open_errno));    
	    errors++;
	}
	
	if (0L == name_resolution_val &&
	    (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 0 == access(system_rc_file,WRITE_ACCESS))) {

	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCheckNameresolution,
			      "Check \"name-resolution\" on RC file: %s"),
		      system_rc_file);
	}

	if (!hostname[0] || !hostdomain[0]) {

	    if (hostfullname[0]) {
		char * p = strchr(hostfullname,'.');
		if (!hostname[0]) 
		    strfcpy(hostname,hostfullname,sizeof hostname);
		if (!hostdomain[0]) {
		    if (p)
			strfcpy(hostdomain,p,sizeof hostdomain);
		    else 	      
			strfcpy(hostdomain,".localdomain",sizeof hostdomain);
		}

	    } else {

		if (0 != (global_mode & READ_FLAG_GLOBAL_ONLY) || 0 == access(system_rc_file,WRITE_ACCESS)) 
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmDefineHostname,
				      "Give \"hostname\", \"hostfullname\" and \"hostdomain\" on RC file: %s"),
			      system_rc_file);

		if (!hostname[0]) 
		    strfcpy(hostname,"localhost",sizeof hostname);
		if (!hostfullname[0]) {
			strfcpy(hostfullname,hostname,sizeof hostfullname);
			/* Hostdomain may be empty, in that case hostfullname
			   is just hostname
			*/
			strfcat(hostfullname,hostdomain,sizeof hostfullname);
		}
		if (!hostdomain[0]) 
		    strfcpy(hostdomain,".localdomain",sizeof hostdomain);		
		errors++;       
	    }
	}   
    }    
}

static void name_resolution_register_if_ok P_((enum name_resolution tag));
static void name_resolution_register_if_ok(tag)
     enum name_resolution tag;
{
    if (0 == name_resolution_errno(tag))
	name_resolution_val |= 1 << tag;

}

/* Returns 1 if success, 0 failure, -1 if expansion is not string */
static int pager_expand_func(kw,ki,keyword,continue_mask)
     struct dt_estr_info *kw;
     struct dt_keyword_info *ki;
     unsigned int keyword;
     unsigned int continue_mask;
{
    char *cp;

    if (0 != (continue_mask & pager_KW_PAGER) &&
	NULL != (cp = getenv("PAGER"))) {

	estr_set_expanded(kw,cp);

	DPRINT(Debug,12,(&Debug, 
			 "pager_expand_func=1: continue_mask=x%04x $PAGER=%s\n",
			 continue_mask,cp));
	
	return 1;
    }

    switch(keyword) {
    case pager_kw_builtin:

	kw->expanded = strmcpy(kw->expanded,"builtin");
	kw->expanded_keyword = keyword;

	DPRINT(Debug,12,(&Debug, 
			 "pager_expand_func=1: keyword=%d expanded=%s\n",
			 keyword,kw->expanded));
		
	return 1;
    }


    return 0;
}

/* Returns 1 if success, 0 failure, -1 if expansion is not string */
static int editor_expand_func(kw,ki,keyword,continue_mask)
     struct dt_estr_info *kw;
     struct dt_keyword_info *ki;
     unsigned int keyword;
     unsigned int continue_mask;
{
    char *cp;

    if (0 != (continue_mask & editor_KW_EDITOR) &&
	NULL != (cp = getenv("EDITOR"))) {

	estr_set_expanded(kw,cp);

	DPRINT(Debug,12,(&Debug, 
			 "editor_expand_func=1: continue_mask=x%04x $EDITOR=%s\n",
			 continue_mask,cp));
	
	return 1;
    }

    if (0 != (continue_mask & editor_KW_VISUAL) &&
	NULL != (cp = getenv("VISUAL"))) {

	estr_set_expanded(kw,cp);

	DPRINT(Debug,12,(&Debug, 
			 "editor_expand_func=1: continue_mask=x%04x $VISUAL=%s\n",
			 continue_mask,cp));
	
	return 1;
    }


    switch(keyword) {
    case editor_kw_builtin:

	kw->expanded = strmcpy(kw->expanded,"builtin");
	kw->expanded_keyword = keyword;

	DPRINT(Debug,12,(&Debug, 
			 "editor_expand_func=1: keyword=%d expanded=%s\n",
			 keyword,kw->expanded));
		
	return 1;
    }


    return 0;
}

/* Returns 1 if success, 0 failure, -1 if expansion is not string */
static int alteditor_expand_func(kw,ki,keyword,continue_mask)
     struct dt_estr_info *kw;
     struct dt_keyword_info *ki;
     unsigned int keyword;
     unsigned int continue_mask;
{

    char *cp;

    if (0 != (continue_mask & alteditor_KW_EDITOR) &&
	NULL != (cp = getenv("EDITOR"))) {

	estr_set_expanded(kw,cp);

	DPRINT(Debug,12,(&Debug, 
			 "alteditor_expand_func=1: continue_mask=x%04x $EDITOR=%s\n",
			 continue_mask,cp));
	
	return 1;
    }

    if (0 != (continue_mask & alteditor_KW_VISUAL) &&
	NULL != (cp = getenv("VISUAL"))) {

	estr_set_expanded(kw,cp);

	DPRINT(Debug,12,(&Debug, 
			 "alteditor_expand_func=1: continue_mask=x%04x $VISUAL=%s\n",
			 continue_mask,cp));
	
	return 1;
    }


    switch(keyword) {
	const char * editor_val;
	unsigned int editor_keyword, editor_mask;

    case alteditor_kw_EDITOR:
	
	editor_val =  give_dt_estr_as_str(&editor_e,"editor",
					  &editor_keyword,
					  &editor_mask);
	
	if (!editor_val)
	    editor_val = "EDITOR";
	
	kw->expanded_keyword = 
	    editor_keyword ? editor_keyword : keyword;

	kw->expanded = strmcpy(kw->expanded,editor_val);
					  
	DPRINT(Debug,12,(&Debug, 
			 "alteditor_expand_func=1: keyword=%d expanded=%s expanded_keyword=%d\n",
			 keyword,kw->expanded,
			 kw->expanded_keyword));
	return 1;
    }

    return 0;
}

char elmrc_charset[SLEN];
static charset_t set_elmrc_charset(filename,errors)
     const char *filename;
     int *errors;
{
    if (elmrc_charset[0]) {
	
	charset_t cs = MIME_name_to_charset(elmrc_charset,0);
	
	if (cs) {
	    const char * csn;
	    
	    csn = get_charset_MIME_name(cs);
	    
	    if (csn) {
		DPRINT(Debug,20,(&Debug,
				 "set_elmrc_charset: %s: Using %s (%s) as charset\n",
				 filename,elmrc_charset,csn));
	    } else {
		DPRINT(Debug,20,(&Debug,
				 "set_elmrc_charset: %s: Using %s as charset\n",
				 filename,elmrc_charset,csn));
	    }
	    
	    
	    if (!(CS_mapping & charset_properties(cs))) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmUnsupportedRCCharset,
				  "Charset %s is unsupported at @charset option in \"%s\" file"),
			  csn ? csn : elmrc_charset,filename);
		
		(*errors)++;
	    }
	    
	    return cs;
	}
	
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmUnsupportedRCCharset,
			  "Charset %s is unsupported at @charset option in \"%s\" file"),
		  elmrc_charset,
		  filename);
	
	(*errors)++;
    }

    /* Return NULL if @charset was not set on this file */

    return NULL;
}


int init_defaults(global_mode) 
     int global_mode;
{
    
    char     *cp;
    FILE * file;

    int rc_open_errno = 0;

    struct parsed_rc *rc = NULL;
    charset_t   rcset = NULL;

    int i;

    /* Establish some defaults in case elmrc is incomplete or not there.
     * Defaults for other elmrc options were established in their
     * declaration - in elm.h.  And defaults for sent_mail and recvd_mail
     * are established after the elmrc is read in since these default
     * are based on the folders directory name, which may be given
     * in the emrc.
   */
    
    DPRINT(Debug,8,(&Debug, 
		    "init_defaults: global_mode = %d%s\n",
		    global_mode,
		    (global_mode & READ_FLAG_GLOBAL_ONLY) ? " READ_FLAG_GLOBAL_ONLY" : ""));

    if (init_defaults_called) {
	DPRINT(Debug,1,(&Debug,"init_defaults called twice!!\n"));	
    }
    init_defaults_called = 1;

    errors = 0;

    for (i = 0; i < nr_COUNT; i++) {
	/* If duplicate call */
	if (name_resolutions[i].result)
	    free(name_resolutions[i].result);
	name_resolutions[i].result = NULL;
	name_resolutions[i].errno_val = 0;

	if (name_resolutions[i].init != nr_null_init &&
	    name_resolutions[i].init != nr_not_avail) {

	    DPRINT(Debug,9,(&Debug, 
			    "init_defaults: Initializing name resolution %s\n",
			    name_resolutions[i].tag));

	    name_resolutions[i].init(& (name_resolutions[i]), i);
	}
 
    }

    if (!quote_prefix)
	quote_prefix = 
	    new_string2(ASCII_SET,s2us("> "));  /* prefix char(s) for msgs          */

    /* Setup default resolutions */
#ifdef GETHOSTNAME
    name_resolution_register_if_ok(nr_gethostname);
#endif

#ifdef DOUNAME
    name_resolution_register_if_ok(nr_uname);
#endif

#ifdef I_NETDB
    name_resolution_register_if_ok(nr_lookup);
#endif

#ifdef USEGETDOMAINNAME
    name_resolution_register_if_ok(nr_getdomain);
#endif
    
#ifdef USEMAILNAME
    name_resolution_register_if_ok(nr_mailname);
#endif

    /* Determine the default mail file name.
     * 
     * First look for an environment variable MAIL, then
     * use then mailhome if it is not found
     */
    
    {
	char *X = elm_message(FRM("%s%s"), mailhome, username);
	set_dt_estr(&defaultfile_e,X,"MAIL",0,0);
	free(X);
    }

    allowed_precedences[0] = '\0';
    /* no defaults for those */


    set_dt_estr(&shell_e,shell0,"SHELL",0,0);
    /* Shell is set from /etc/passwd in user_init () */

#ifdef USE_BUILTIN_PAGER
    set_dt_estr(&pager_e, "builtin++", NULL, pager_kw_builtin,0);
#else
    set_dt_estr(&pager_e,default_pager,NULL, 0,pager_KW_PAGER);
#endif    

    set_dt_estr(&temp_dir_e,default_temp,"TMPDIR",0,0);
    set_dt_estr(&editor_e,default_editor,NULL,0,
		editor_KW_EDITOR|editor_KW_VISUAL);
    set_dt_estr(&alternative_editor_e,"EDITOR",NULL,
		alteditor_kw_EDITOR,
		alteditor_KW_EDITOR|alteditor_KW_VISUAL);

    if ((cp = getenv("MAILCAPS")) == NULL) {
	/* RFC 1524:

   For UNIX, a path search of mailcap files is specified.  The default
   path search is specified as including at least the following:

   $HOME/.mailcap:/etc/mailcap:/usr/etc/mailcap:/usr/local/etc/mailcap

	*/
	char *X[] = { "/etc/mailcap", "/usr/etc/mailcap",
		      "/usr/local/etc/mailcap", NULL };
	char *G = elm_message(FRM("%s/.mailcap"),home);
	char **X1;
	
	/* Fake recursive call -- expand_dt_path ignores non-existiant
	   entries.
	 */
	DPRINT(Debug,9,(&Debug,
			"Adding default entry to metamail-mailcaps: %s\n",
			G));
	expand_dt_path(&metamail_mailcaps,"metamail-mailcaps",
		       G,2);
	for (X1 = X; *X1; X1++) {
	    DPRINT(Debug,9,(&Debug,
			    "Adding default entry to metamail-mailcap: %s\n",
			    *X1));
	    expand_dt_path(&metamail_mailcaps,"metamail-mailcaps",
			   *X1,2);	    
	}


	DPRINT(Debug,9,(&Debug,
			"Adding default entry to internal-mailcaps: %s\n",
			G));
	expand_dt_path(&internal_mailcaps,"internal-mailcaps",
		       G,2);
	for (X1 = X; *X1; X1++) {
	    DPRINT(Debug,9,(&Debug,
			    "Adding default entry to internal-mailcap: %s\n",
			    *X1));
	    expand_dt_path(&internal_mailcaps,"internal-mailcaps",
			   *X1,2);	    
	}
	
	free(G);
    } else {
	DPRINT(Debug,9,(&Debug,
			"Adding unexpanded entry to metamail-mailcaps: $MAILCAPS\n"
			));

	metamail_mailcaps.unexpanded = safe_strdup("$MAILCAPS");


	DPRINT(Debug,9,(&Debug,
			"Adding unexpanded entry to internal-mailcaps: $MAILCAPS\n"
			));

	internal_mailcaps.unexpanded = safe_strdup("$MAILCAPS");
    }

    folders_e.unexpanded    = elm_message(FRM("~/%s"), default_folders);
    map_txtdir_e.unexpanded = safe_strdup("{lib}/elm.map.txt");
    map_bindir_e.unexpanded = safe_strdup("{lib}/elm.map.bin");

    readmsg_e.unexpanded = safe_strdup("{bin}/readmsg");
       
    calendar_file_e.unexpanded = 
	elm_message(FRM("~/%s"), dflt_calendar_file);

    dead_letter_dir_e.unexpanded = safe_strdup("~/Canceled.mail.dir");

    strfcpy(e_editor, emacs_editor, sizeof e_editor);
    strfcpy(v_editor, default_editor, sizeof v_editor);

#ifdef USE_PGP
    set_dt_estr(&gpg_path_e,GPG_PATH,NULL,0,0);
    set_dt_estr(&pgp2_path_e,PGP2_PATH,NULL,0,0);
    set_dt_estr(&pgp5_dir_e,PGP5_DIR,NULL,0,0);
#endif
    set_dt_estr(&metamail_path_e,METAMAIL_PATH,NULL,0,0);            
    set_dt_estr(&printout_e,DEFAULT_PRINTOUT,NULL,0,0);            

    {
	char *X[] = { "/usr/bin/wine", "wine", "/bin/sh", "sh", NULL };

	/* However do not add "true" or "false" to blacklist. */

	char *ignore[] = { "/bin/false", "false", "/bin/true", 
			   "true", NULL };

	char **X1;
	char *Y;
	
	/* Fake recursive call -- expand_dt_path ignores non-existiant
	   entries.
	 */

	for (X1 = X; *X1; X1++) {
	    DPRINT(Debug,9,(&Debug,
			    "Adding default entry to mailcap-blacklist-programs: %s\n",
			    *X1));
	    expand_dt_path(&mailcap_bl_programs,"mailcap-blacklist-programs",
			   *X1,2);	    
	}

#ifdef GETUSERSHELL
	/* Also include shells from /etc/shells  */

	while (NULL != (Y = getusershell())) {
	    char * Y1;
	    for (X1 = X; *X1; X1++) {
		if (0 == strcmp(*X1,Y))
		    break;
	    }

	    if (!*X1) {
		for (X1 = ignore; *X1; X1++) {
		    if (0 == strcmp(*X1,Y))
			break;
		}
	    }

	    /* And if not in default list */
	    if (!*X1) {
		DPRINT(Debug,9,(&Debug,
				"Adding default entry to mailcap-blacklist-programs: %s\n",
				Y));
		expand_dt_path(&mailcap_bl_programs,"mailcap-blacklist-programs",
			       Y,2);	    


	    }

	    Y1 = strrchr(Y,'/');
	    if (Y1) {
		Y1++;

		for (X1 = X; *X1; X1++) {
		    if (0 == strcmp(*X1,Y1))
			break;
		}

		if (!*X1) {
		    for (X1 = ignore; *X1; X1++) {
			if (0 == strcmp(*X1,Y1))
			    break;
		    }
		}
				
		/* And if not in default list */
		if (!*X1) {
		    DPRINT(Debug,9,(&Debug,
				    "Adding default entry to mailcap-blacklist-programs: %s\n",
				    Y1));
		    expand_dt_path(&mailcap_bl_programs,"mailcap-blacklist-programs",
				   Y1,2);	    
		    		    
		}
	    }
	}
	endusershell();
#endif
    }

    elmrc_charset[0] = '\0';   /* Reset old @charset value */

    /* try system-wide rc file */
    file = fopen(system_rc_file, "r");
    if ( file != NULL ) {

	/* This process only options marked with FL_ASCIIFIRST */

	rc = do_rc_parse(file, SYSTEM_RC,system_rc_file,&errors);

	fclose(file);
    } else {
	rc_open_errno = errno;

	DPRINT(Debug,1,(&Debug,
			"Opening %s failed:  (errno=%d) %s\n",
			system_rc_file,rc_open_errno,strerror(rc_open_errno)));
    }
       
    /* hostdomain, hostfullname, hostname, name-resolution, mailname
       are set with FL_ASCIIFIRST so they are set when setup_names() is called
    */
    setup_names(rc_open_errno, global_mode);

    /* Also unidata  ... notice temporary variable */
    expand_map_bin("unidata", unidata_path_global, raw_unidata_path,
		   sizeof(unidata_path));

    /* Also bindata  ... notice temporary variable */
    expand_map_bin("bindata", bindata_path_global, raw_bindata_path,
		   sizeof(bindata_path));

    parse_iso2022_map(system_iso2022_sets,&errors,ml_system);
    system_charset_map = load_charset_map(system_mime_charsets,&errors);

    map_system_charset(display_locale,display_codeset,1);

    /* Returns NULL, if @charset is not set */
    rcset = set_elmrc_charset(system_rc_file,&errors);
    
    if (rc) {    /* Process most of  elm.rc options */
	do_rc_process(rc, SYSTEM_RC,system_rc_file,rcset,&errors);
	free_parsed_rc(&rc);
    } else if (!rc_open_errno)
	errors++;

#ifdef REMOTE_MBX
    parse_service_entries(system_mail_services,1 /* SYSTEM */,&errors);
#endif

    for (i = 0; i < hooklist_len; i++) {
	hooklist[i](&errors);
    }

    DPRINT(Debug,3,(&Debug, 
		    "init_defaults=%d (errors)\n",errors));

    return errors;
}

enum charset_keyword special_charset_keyword(value)
     const char *value;
{
    if (0 == strcmp(value,"SYSTEM"))
	return cskw_SYSTEM;
    else if (0 == strcmp(value,"DISPLAY"))
	return cskw_DISPLAY;
    else if (0 == strcmp(value,"TEXT"))
	return cskw_TEXT;
    else
	return cskw_none;
}

void post_process_charset_options()
{
    switch (special_charset_keyword(raw_text_charset)) {
    case cskw_SYSTEM:   text_charset = system_charset;  break;
    case cskw_DISPLAY:  text_charset = display_charset; break;
    case cskw_TEXT:     
	text_charset = display_charset;
	strfcpy(raw_text_charset,"DISPLAY",sizeof raw_text_charset);
	break;
    default:
	break;
    }

    switch (special_charset_keyword(raw_display_charset)) {
    case cskw_SYSTEM:   wanted_display_charset = system_charset;  break;
    case cskw_DISPLAY:  wanted_display_charset = display_charset; break;
    case cskw_TEXT:     wanted_display_charset = text_charset;    break;
    default:                                                      break;
    }

    if (!wanted_display_charset ||
	0 == (charset_properties(wanted_display_charset) & CS_mapping)) {
	if (0 != charset_properties(system_charset)) 
	    wanted_display_charset = system_charset;
	else
	    wanted_display_charset = ASCII_SET;

	DPRINT(Debug,2,(&Debug,
			"Setting (wanted) display-charset to %s instead.\n",
			wanted_display_charset->MIME_name ?
			wanted_display_charset->MIME_name :
			"<no MIME name>"));
    }
    if (!text_charset || !text_charset->MIME_name) {
	text_charset = ASCII_SET;
	DPRINT(Debug,2,(&Debug,
			"Setting text-charset to %s instead.\n",
			text_charset->MIME_name ?
			text_charset->MIME_name :
			"<no MIME name"));
    }

    switch (special_charset_keyword(raw_local_fs_charset)) {
    case cskw_SYSTEM:   local_fs_charset = system_charset;  break;
    case cskw_DISPLAY:  local_fs_charset = display_charset; break;
    case cskw_TEXT:     local_fs_charset = text_charset;    break;
    default:                                                break;
    }

    switch (special_charset_keyword(raw_default_nomime_charset)) {
    case cskw_SYSTEM:   default_nomime_charset = system_charset;  break;
    case cskw_DISPLAY:  default_nomime_charset = display_charset; break;
    case cskw_TEXT:     default_nomime_charset = text_charset;    break;
    default:                                                      break;
    }

    switch (special_charset_keyword(raw_default_mimetext_charset)) {
    case cskw_SYSTEM:   default_mimetext_charset = system_charset;  break;
    case cskw_DISPLAY:  default_mimetext_charset = display_charset; break;
    case cskw_TEXT:     default_mimetext_charset = text_charset;    break;
    default:                                                        break;
    }

#ifdef REMOTE_MBX
    switch (special_charset_keyword(raw_imap_charset)) {
    case cskw_SYSTEM:   imap_charset = system_charset;  break;
    case cskw_DISPLAY:  imap_charset = display_charset; break;
    case cskw_TEXT:     imap_charset = text_charset;    break;
    default:                                            break;
    }
#endif
	
}


static init_default_hook **hooklist_rc = NULL;
static int                hooklist_rc_len = 0;

static int read_rc_file_called = 0;

void add_read_rc_file_hook(hook)
     init_default_hook *hook;
{

    if (read_rc_file_called) {
	hook(&errors);
    }

    hooklist_rc = safe_realloc(hooklist_rc,
			       (hooklist_rc_len+1) * sizeof(hooklist_rc[0]));

    hooklist_rc[hooklist_rc_len] = hook;
    hooklist_rc_len++;
}



int read_rc_file(flag)
     int flag;
{
    /** this routine does all the actual work of reading in the
	.rc file... **/

    int i;
    struct parsed_rc *rc = NULL;
    int open_error = 0;
    charset_t   rcset = NULL;

    errors = 0;
    
    if (read_rc_file_called) {
	DPRINT(Debug,1,(&Debug,"read_rc_file called twice!\n"));
    }
    read_rc_file_called = 1;

    elmrc_charset[0] = '\0';   /* Reset old value */

    /* Look for the elmrc file */
    open_error = can_open(user_rc_file, "r");
    if (open_error) {
    
	if (user_rc_set) 
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedUserRC,
			      "Failed to open RC file %s: %s"),
		      user_rc_file,strerror(open_error));    

	DPRINT(Debug,2,(&Debug,
			"Warning:User has no \"%s\" file (can_open)\n\n",
			user_rc_file));

    } else {
       
	FILE *file = NULL;
    
	if ((file = fopen(user_rc_file, "r")) == NULL) {
	    open_error = errno;

	    if (user_rc_set) 
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedUserRC,
				  "Failed to open RC file %s: %s"),
			  user_rc_file,strerror(open_error));    


	    DPRINT(Debug,2,(&Debug,
			    "Warning:Open \"%s\" failed\n\n",
			user_rc_file));

	} else {

	    /* This process only options marked with FL_ASCIIFIRST */

	    rc = do_rc_parse(file, LOCAL_RC, 
			     user_rc_set ? user_rc_file: ELMRCFILE,
			     &errors);

	    fclose(file);
	}
    }


    /* unidata may be needed ... */
    expand_map_bin("unidata", unidata_path, raw_unidata_path,
		   sizeof(unidata_path));

    /* Also bindata  ... */
    expand_map_bin("bindata", bindata_path, raw_bindata_path,
		   sizeof(bindata_path));

    parse_iso2022_map(user_iso2022_sets,&errors, ml_user);
    user_charset_map   = load_charset_map(user_mime_charsets,&errors);

    rcset = set_elmrc_charset(user_rc_set ? user_rc_file: ELMRCFILE,
			      &errors);

    if (rc) { /* Process most of  elm.rc options */
	do_rc_process(rc,  LOCAL_RC, 
		      user_rc_set ? user_rc_file: ELMRCFILE,
		      rcset,&errors);
	free_parsed_rc(&rc);
    } else if (!open_error)
	errors++;

#ifdef REMOTE_MBX
    parse_service_entries(user_mail_services,0 /* USER */,&errors);
#endif 

    for (i = 0; i < hooklist_rc_len; i++) {
	hooklist_rc[i](&errors);
    }


    post_init_check(flag);
    
    DPRINT(Debug,3,(&Debug, 
		    "read_rc_file=%d (errors)\n",errors));

    return errors;
}

void post_init_check(flag) 
     int flag; 
{

    struct stat st_buffer;
    const char *t, *t1;

    t  = give_dt_estr_as_str(&folders_e,"maildir",NULL,NULL);
    t1 = give_dt_estr_as_str(&extra_mailbox_dir_e,
			     "extra-mailbox-dir",NULL,NULL);
       
    if (t1 && t && 0 == strcmp(t1,t)) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmExtraNoFolders,
			  "\"extra-mailbox-dir\" should not be \"folders\" or \"maildir\" directory."));	    

	free(extra_mailbox_dir_e.unexpanded);
	extra_mailbox_dir_e.unexpanded = NULL;

	mark_local_changed(& extra_mailbox_dir_e);
    }

    t = give_dt_estr_as_str(&dead_letter_dir_e,"dead-letter-dir",NULL,NULL);
			    
    /* dead-letter-dir is create automatically, if it does not exists */
    if (t &&
	0 == stat(t,&st_buffer) &&
#ifdef S_ISDIR
	!S_ISDIR(st_buffer.st_mode)
#else
	S_IFDIR != (st_buffer.st_mode & S_IFMT)
#endif
	) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadDeadLetterDir,
			  "\"dead-letter-dir\" should be directory."));

	free(dead_letter_dir_e.unexpanded);
	dead_letter_dir_e.unexpanded = NULL;

	mark_local_changed(& dead_letter_dir_e);
    }
    
    if (! attachment_dir_e.unexpanded) {
	char * tmp = elm_message(FRM("%s/Documents"),home);
	
	if (0 == access(tmp,ACCESS_EXISTS)) {
	    attachment_dir_e.expanded = tmp;
	    attachment_dir_e.unexpanded = safe_strdup("~/Documents");

	} else
	    free(tmp);
    } 
                      
    if ((elm_timeout != 0) && (elm_timeout < 10)) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTimeoutLTTen,
			  "Warning: timeout is set to less than 10 seconds. Ignored."));
	elm_timeout = 0;
    }
    
    if (readdatapercentinc < 1) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmReadDataIncrement,
			  "Warning: readdatapercentinc is set to less than 1.  Ignored."));
	readdatapercentinc = 1;

	mark_local_changed(& readdatapercentinc);
    }

    if (readmsginc < 1) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmReadMessageIncrement,
			  "Warning: readmsginc is set to less than 1.  Ignored."));
	readmsginc = 1;

	mark_local_changed(& readmsginc);
    }

    if (sleepmsg < 0) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSleepMessageInvalid,
			  "Warning: sleepmsg is set to less than 0.  Setting to 0."));
	sleepmsg = 0;

	mark_local_changed(& sleepmsg);
    }
    
#ifdef BACKGROUD_PROCESSES       
    if (background_wait_time  < 0) {
	errors++;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBackgroundLLT,
			  "Warning: background_wait_time is set less than 0. Setting to 0."));
	background_wait_time = 0;

	mark_local_changed(& background_wait_time); 
    }
#endif
    
    
    /* If recvd_mail or sent_mail havent't yet been established in
     * the elmrc, establish them from their defaults.
     * Then if they begin with a metacharacter, replace it with the
     * folders directory name.
     */
    if(!recvd_mail_e.unexpanded) {
	recvd_mail_e.unexpanded = safe_strdup(default_recvdmail);
    }
    
    if(!sent_mail_e.unexpanded) {
	sent_mail_e.unexpanded  = safe_strdup(default_sentmail);
    }

    if (-1 == auto_copy_sent) {
	if (dt_estr_is_disabled(&sent_mail_e)) {
	    DPRINT(Debug,5,(&Debug,"'copy' defaults to OFF (sentmail is disabled)\n"));
	    auto_copy_sent = 0;
	} else {
	    DPRINT(Debug,5,(&Debug,"'copy' defaults to ON\n"));
	    auto_copy_sent = 1;
	}
    }


    expand_map_bin("unidata", unidata_path, raw_unidata_path,
		   sizeof(unidata_path));

    expand_map_bin("bindata", bindata_path, raw_bindata_path,
		   sizeof(bindata_path));
    
    map_system_charset(display_locale,display_codeset,0);

       
#if defined(WCHAR) && defined(__STDC_ISO_10646__)
    change_system_charset();
#endif 

    /* Current display charset is locale charset */
    display_charset = system_charset;
    
    if (!special_charset_keyword(raw_text_charset)) {

	if (! load_charset_map_info(&text_charset,
				    raw_text_charset) ) {

	    char buffer[LONG_STRING];

	    do_expand_env("text-charset", buffer, 
			  raw_text_charset, 
			  sizeof(buffer));
	    text_charset = MIME_name_to_charset(buffer,CHARSET_create);
	}

	if (!(CS_mapping & charset_properties(text_charset))) {
	    errors++;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmUnknownTextCharset,
			      "text-charset %s is unknown, conversion to text-charset does not work"),
		      text_charset->MIME_name ? 
		      text_charset->MIME_name :
		      raw_text_charset);
	    strfcpy(raw_text_charset,"DISPLAY",sizeof raw_text_charset);
	    text_charset = display_charset;

	    mark_local_changed(& raw_text_charset);
	}        
    } 
    
    if (!special_charset_keyword(raw_display_charset)) {

	if (! load_charset_map_info(& wanted_display_charset,
				    raw_display_charset )) {

	    char buffer[LONG_STRING];
	    do_expand_env("display-charset", buffer, 
			  raw_display_charset, 
			  sizeof(buffer));
	    wanted_display_charset = MIME_name_to_charset(buffer,CHARSET_create);
	}

	if (!(CS_mapping & charset_properties(wanted_display_charset))) {
	    errors++;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnknownDisplayCharset,
			      "display-charset %s is unsupported, charset switching does not work"),
		      wanted_display_charset->MIME_name ? 
		      wanted_display_charset->MIME_name :
		      raw_display_charset);
	    strfcpy(raw_display_charset,"SYSTEM",
		    sizeof raw_display_charset);
	    wanted_display_charset = system_charset;

	    mark_local_changed(& raw_display_charset);
	}
    }

    if (!special_charset_keyword(raw_local_fs_charset)) {

	if (! load_charset_map_info(&local_fs_charset,
				    raw_local_fs_charset) ) {
	    char buffer[LONG_STRING];
	    do_expand_env("local-fs-charset", buffer, 
			  raw_local_fs_charset, 
			  sizeof(buffer));
	    local_fs_charset = MIME_name_to_charset(buffer,CHARSET_create);
	}

	if (!(CS_mapping & charset_properties(local_fs_charset))) {
	    errors++;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnknownFSCharset,
			      "local-fs-charset %s is unsupported, file access does not work"),
		      local_fs_charset->MIME_name ? 
		      local_fs_charset->MIME_name :
		      raw_local_fs_charset);
	    strfcpy(raw_local_fs_charset,"SYSTEM",
		    sizeof raw_local_fs_charset);
	    local_fs_charset = system_charset;		

	    mark_local_changed(& raw_local_fs_charset);
	}
    }
    
    if (!special_charset_keyword(raw_default_nomime_charset)) {

	if (! load_charset_map_info(& default_nomime_charset,
				    raw_default_nomime_charset) ) {
	    char buffer[LONG_STRING];

	    do_expand_env("default-nomime-charset", buffer, 
			  raw_default_nomime_charset, 
			  sizeof(buffer));
	    default_nomime_charset = MIME_name_to_charset(buffer,CHARSET_create);
	}

	if (!(CS_mapping & charset_properties(default_nomime_charset))) {
	    errors++;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnknownDef1Charset,
			      "default-nomime-charset %s is unsupported"),
		      default_nomime_charset->MIME_name ? 
		      default_nomime_charset->MIME_name :
		      raw_default_nomime_charset);

		strfcpy(raw_default_nomime_charset,"SYSTEM",
			sizeof raw_default_nomime_charset);
		default_nomime_charset = system_charset;		

		mark_local_changed(&  raw_default_nomime_charset);
	}
    }

    if (!special_charset_keyword(raw_default_mimetext_charset)) {

	if (! load_charset_map_info(& default_mimetext_charset,
				    raw_default_mimetext_charset) ) {

	    char buffer[LONG_STRING];
	    do_expand_env("default-mime-text-charset", buffer, 
			  raw_default_mimetext_charset, 
			  sizeof(buffer));
	    default_mimetext_charset = MIME_name_to_charset(buffer,CHARSET_create);
	}

	if (!(CS_mapping & charset_properties(default_mimetext_charset))) {
	    errors++;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnknownDef2Charset,
			      "default-mime-text-charset %s is unsupported"),
		      default_mimetext_charset->MIME_name ? 
		      default_mimetext_charset->MIME_name :
		      raw_default_mimetext_charset);

		strfcpy(raw_default_mimetext_charset,"US-ASCII",
			sizeof raw_default_mimetext_charset);
		default_mimetext_charset = ASCII_SET;		

		mark_local_changed(&  raw_default_mimetext_charset);
	}
    }

    
#ifdef REMOTE_MBX
    if (!special_charset_keyword(raw_imap_charset)) {

	if (! load_charset_map_info(& imap_charset,
				    raw_imap_charset) ) {

	    char buffer[LONG_STRING];
	    do_expand_env("imap-charset", buffer, 
			  raw_imap_charset, 
			  sizeof(buffer));
	    imap_charset = MIME_name_to_charset(buffer,CHARSET_create);
	}

	if (!(CS_mapping & charset_properties(imap_charset))) {
	    errors++;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnknownIMAPCharset,
			      "imap-charset %s is unsupported"),
		      imap_charset->MIME_name ? 
		      imap_charset->MIME_name :
		      raw_imap_charset);

	    strfcpy(raw_imap_charset,"SYSTEM",
		    sizeof raw_imap_charset);
	    imap_charset = system_charset;		

	    mark_local_changed(&  raw_imap_charset);
	}
    }
#endif
    
    post_process_charset_options();
    
    if (!(CS_mapping & charset_properties(system_charset))) {
	/* Don't increment errors variable because this comes
	 * from locale (via mime.charsets) and not from elmrc
	 */
	
	if (system_charset->MIME_name)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnsupportedLocale,
			      "WARNING: Locale %s (charset %s) is unsupported, will cause problems!"), 
		      display_locale,system_charset->MIME_name);
	else
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmUnsupportedLocale2,
			      "WARNING: Locale %s is unsupported, will cause problems!"), 
		      display_locale);
	locale_error = 1;
    } 
    
    if (locale_error == 1) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmLocaleProblem,
			  "Problem with locale (system character set)! Elm ME+ will behave erratically."));
	/* Long wait that message will be noticed ... */
#if POLL_METHOD	  
	wait_for_timeout(10);
#else
	sleep(10);
#endif	    
	locale_error++;
    }
    
    if (
	(
	 (wanted_display_charset != system_charset &&	     
	  0 != (CS_mapping & charset_properties(wanted_display_charset)) &&
	  (!wanted_display_charset->MIME_name ||
	   0 != istrcmp(wanted_display_charset->MIME_name,"US-ASCII")) 
#ifndef ASCII_CTYPE
	  && (!wanted_display_charset->MIME_name ||
	      0 != istrcmp(wanted_display_charset->MIME_name,"ISO-8859-1"))
#endif
	  )
	 || allow_charset_switching
	 ) &&
	0 != strcmp(unidata_path,"none")) {
	
	/* We are going to need unidata information --
	   therefore check availivility ! */
	
	if (0 == (flag & READ_FLAG_UNIDATA_INIT) 
	    &&
	    !default_unidata()) {
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet,ElmNoUnicodeData,
			      "No Unicode data available: unidata: %s"),
		      raw_unidata_path);
	    
	    errors++;
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet,ElmUnidataSetting,
			      "Setting unidata=none, displaycharset=SYSTEM, allow-charset-switching=OFF"));
	    
	    strfcpy(raw_unidata_path,"none",sizeof raw_unidata_path);
	    strfcpy(unidata_path,"none",sizeof unidata_path);
	    
	    strfcpy(raw_display_charset,"SYSTEM", 
		    sizeof raw_display_charset);
	    wanted_display_charset = system_charset;		

	    allow_charset_switching = 0;

	    mark_local_changed(& raw_unidata_path);
	    mark_local_changed(& raw_display_charset);
	    mark_local_changed(& allow_charset_switching);

	}
	
    }
    
    if (0 != strcmp(bindata_path,"none") &&
	0 != strcmp(bindata_path_global,bindata_path)) {

	if (0 == (flag & READ_FLAG_BINDATA_INIT) 
	    &&
	    !default_bindata()) {
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet,ElmBinaryMapData,
			      "No binary map data available: bindata: %s"),
		      raw_bindata_path);
	    
	    errors++;

	    strfcpy(raw_bindata_path,"none",sizeof raw_bindata_path);
	    strfcpy(bindata_path,"none",sizeof bindata_path);

	    mark_local_changed(& raw_bindata_path);
	}
    }

#ifdef USE_PGP
    if (pgp_none == give_dt_enumerate_as_int(&send_pgp_version)) {
	const char * gpg_path_val  = give_dt_estr_as_str(&gpg_path_e, "gpg", 
							 NULL,NULL);
	const char * pgp2_path_val = give_dt_estr_as_str(&pgp2_path_e, "pgp2",
							 NULL,NULL);
	const char * pgp5_dir_val  = give_dt_estr_as_str(&pgp5_dir_e, 
							 "pgp5-dir",NULL,NULL);

	if (gpg_path_val &&  '/' == gpg_path_val[0]) 
	    send_pgp_version.val = gpg;
	if (pgp5_dir_val && '/' == pgp5_dir_val[0]) 
	    send_pgp_version.val = pgp5;
	if (pgp2_path_val && '/' == pgp2_path_val[0]) 
	    send_pgp_version.val = pgp2;
    }
#endif
    
    if (DEFAULT_SORT_ORDER == fb_sort_order.val) {
	fb_sort_order.val = local_dir_sortby.val;

	DPRINT(Debug,10,(&Debug, 
			 "file-browser-sortby = %d (from local-dir-sortby)\n",
			 fb_sort_order.val));

    }

    if (-1 == fb_show_title) {
	fb_show_title = user_level.val == user_level_beginner;

	DPRINT(Debug,10,(&Debug, 
			 "file-browser-show-title = %d (userlevel = %d)\n",
			 fb_show_title, user_level.val));
    }


#ifdef USE_DLOPEN
    post_init_shared_options(&errors);
#endif
}



/* Return rest */

char * rc_split_keyword(buffer,line_len,keyword,size,tag,
			    assig,restlen)
     char *buffer; 
     int line_len;				
     char *keyword; 
     int size;
     char **tag;
     int *assig;
     int *restlen;
{
    int i;
    char * c;

    DPRINT(Debug,10,(&Debug,
		     "rc_split_keyword len=%d, buffer=%s\n",
		     line_len,buffer));

    *assig = 0;

    i = 0;
    for (c = buffer; c - buffer < line_len && i < size; c++) {
	if ('_' == *c)
	    keyword[i++] = '-';

	/* allow @charset */
	else if (0 == i && '@' == *c)
	    keyword[i++] = '@';

	else if (!*tag && ':' == *c) {
	    /* Mark tag from buffer */
	    *c = '\0';
	    *tag = buffer;
	    i = 0;
	} else if (!isascii(*c))
	    break;
	else if (!ok_rc_char(*c))
	    break;
	else {
	    keyword[i++] = tolower((unsigned char)*c);
	}
    }

    keyword[i] = '\0';
	
    while (whitespace(*c)) 
	c++;
    
    if ('=' == *c) {
	c++;
	*assig = 1;
    }
    while (whitespace(*c)) 
	c++;

    DPRINT(Debug,10,(&Debug,
		     "rc_split_keyword: keyword(i=%d)=%s\n",
		     i,keyword));

    if (*tag) {
	DPRINT(Debug,10,(&Debug,
			 "rc_split_keyword: tag=%s\n",*tag));	
    }

    *restlen = line_len - (c - buffer);

    DPRINT(Debug,10,(&Debug,
		     "rc_split_keyword=%s, *restlen=%d\n",c,*restlen));	

    return c;
}

/*
 * set the named parameter according to save_info structure.
 * This routine may call itself (DT_SYN or DT_MLT).
 * Also tags params that were set in "local" (personal) RC file
 * so we know to save them back out in "o)ptions" screen.
 * Uses an internal table to decode sort-by params...should be coupled
 * with sort_name(), etc...but...
 */

/* read_rc.c. -- return NULL on failure 
                 otherwise pointer to option
 */


struct rc_save_info_rec *rc_locate_option(rc_options,rc_option_count,
					  word1,negate)
     struct rc_save_info_rec * rc_options;
     int                       rc_option_count;
     char *word1; 
     int *negate;
{
    int x, y = 0;

    if ('!' == *word1) {
	(*negate)++;
	word1++;
    }

    for (x=0; x < rc_option_count; ++x) {
	y = strcmp(word1, rc_options[x].name);
	if (y <= 0) 
	    break;
    }

    if (y != 0) {

	DPRINT(Debug,35,(&Debug,"rc_locate_option: option = %s, not found\n",
			 word1));
	return NULL;
    }

    DPRINT(Debug,35,(&Debug,"rc_locate_option: option = %s, found [%d]\n",
		     word1,x));

    return & (rc_options[x]);
}

int rc_eval_tail(ret,lcl,value,lineno,filename,negate)
     struct rc_save_info_rec *ret;
     int lcl;
     struct string *value;
     int lineno; 
     const char *filename; 
     int negate;
{

    if (lcl == LOCAL_RC)
	ret->flags |= FL_LOCAL;
    ret->flags |= FL_CHANGED;
    
    if (RCTYPE_magic != ret->dt_type->magic)
	panic("RC PANIC",__FILE__,__LINE__,"rc_eval_tail",
	      "Bad config item type",0);
    
    DPRINT(Debug,10,(&Debug,"%s:%d: [line] %s=%S\n",
		     filename,lineno,ret->name,value));

    if (! ret->dt_type->parse_line(ret,
				   lcl,
				   value,
				   lineno,
				   filename,
				   negate))
	return 0;

    return 1;
}

struct rc_save_info_rec * do_set(lcl, word1, word2, filename)
     int lcl;
     char *word1;
     struct string * word2;
     const char *filename;
{
    int negate = 0;

    struct rc_save_info_rec * ret =  rc_locate_option(save_info,
						   NUMBER_OF_SAVEABLE_OPTIONS,
						   word1,
						   &negate);

    if (!ret) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadKeyInElmrc,
			  "I can't understand keyword \"%s\" in line %d in \"%s\" file"),
		  word1, lineno,filename);
	
	return NULL;
    }

    
    if (!rc_eval_tail(ret,lcl,word2,lineno,filename,negate))
	return NULL;

    return ret;
}

static const char WEED_NO_DEFAULT[] = "*clear-weed-list*";
	

int matches_weedlist(buffer)
     char *buffer;
{
	/** returns true iff the first 'n' characters of 'buffer' 
	    match an entry of the weedlist **/
	
	int i;

	static char *default_list[] = 
	    { ">From", "In-Reply-To:",
	      "References:", "Newsgroups:", "Received:",
	      "Apparently-To:", "Message-Id:", "Content-Type:",
	      "Content-Length:", "MIME-Version:",
	      "Content-Transfer-Encoding:",
	      "Content-Language:", "Content-Disposition:",
	      "From", "X-Mailer:", "User-Agent:", "Status:",
	      "X-ELM-",
	      "X-UIDL", /* Generated by some POP deamons */ 
	      "Return-Path:",
	      "List-ID:", 
	      "List-Help:", "List-Subscribe:", "List-Unsubscribe:",
	      "List-Post:", "List-Owner:", "List-Archive:",

	      "Precedence:", "Errors-To:", 
	      
	      /* Microsoft Exchange: */
	      "Thread-Topic:", "Thread-Index:", "X-MS-Has-Attach:",
	      "X-MS-TNEF-Correlator:",

	      /* Sendmail and Elm ME+ 2.5 */
	      "X-MIME-Autoconverted:",

	      /* Postfix */
	      "Delivered-To:",

	      "*end-of-defaults*", NULL
	    };

	const char **list = give_dt_path_as_elems(&weedlist,"weedout");

	/* If fist elem is "*clear-weed-list*" then default list is not 
	   used
	*/

	if (!list || !list[0] ||
	    0 != strcmp(list[0],WEED_NO_DEFAULT)) {
	    

	    for (i = 0; default_list[i]; i++) {
		if (strincmp(buffer, default_list[i], strlen(default_list[i])) == 0) 
		    return(1);

	    }
	    
	}

	if (list) {
	    for (i=0;list[i]; i++) {
		if (strincmp(buffer, list[i], strlen(list[i])) == 0) 
		    return(1);
	    }
	}
	return(0);
}

int is_it_on(word)
    struct string *word;
{
    int r = 0;

    /** Returns TRUE if the specified word is either 'ON', 'YES'
	or 'TRUE', and FALSE otherwise.   
    **/

    r = string_matches_ascii(word,s2us("on"),  SMA_ignore_case|SMA_match_prefix) ||
	string_matches_ascii(word,s2us("yes"), SMA_ignore_case|SMA_match_prefix) ||
	string_matches_ascii(word,s2us("true"),SMA_ignore_case|SMA_match_prefix);

    DPRINT(Debug,20,(&Debug, "is_it_on(%S)=%d\n",word,r));

    return r;
}

char * mode_to_str(mode)
     int mode;
{
	static char modestr[9+1];	/* "rwxrwxrwx\0" */
	char *s = &modestr[0];
	char *modecharp = "rwxrwxrwx";
	int modebit = 0400;

	while (modebit != 0) {
	    if (mode & modebit)
	        *s++ = *modecharp;
	    else
		*s++ = '-';
	    modecharp++;
	    modebit >>= 1;
	}
	*s = '\0';
	return(modestr);
}

int null_option_func(value,enter, lineno, filename)
     char **value; 
     int enter;
     int lineno; 
     const char *filename;
{
    return 0;
}

void free_rc_options() 
{
    int i;

    if (nameres_return_buffer)
	free(nameres_return_buffer);
    nameres_return_buffer = NULL;

    for (i = 0; i < nr_COUNT; i++) {

	if (name_resolutions[i].result)
	    free(name_resolutions[i].result);
	name_resolutions[i].result = NULL;
    }

    for (i = 0; i < NUMBER_OF_SAVEABLE_OPTIONS; i++) {

	if (RCTYPE_magic != save_info[i].dt_type->magic) 
	    panic("RC PANIC",__FILE__,__LINE__,"free_rc_options",
		  "Bad config item type",0);
	    
	DPRINT(Debug,35,(&Debug,
			 "free_rc_options: Freeing save_info[%d] %p %s\n",
			 i, & (save_info[i]), save_info[i].name));

	save_info[i].dt_type->free_option(& (save_info[i]));
    }

    free_delayed_rc_options();

#ifdef USE_DLOPEN
    free_shared_rc_options();
#endif
}



/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
